#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Callable, Iterator, Optional, Protocol, Sequence, TypeVar, cast

from jsonbender import bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    Change,
    IndentingPrinter,
    T,
    associate_by_key,
    is_different_ignoring_order,
    is_unset,
    multi_associate_by_key,
    patch_to_other,
    style,
    write_patch_object_as_json,
)

MT = TypeVar("MT", bound="ModelObject")


class FailureType(Enum):
    INFO = 1
    WARNING = 2
    ERROR = 3


@dataclasses.dataclass
class ValidationContext(object):
    root_object: Any
    template_dir: str
    validation_failures: list[tuple[FailureType, str]] = dataclasses.field(default_factory=list)

    def add_failure(self, failure_type: FailureType, message: str):
        self.validation_failures.append((failure_type, message))

    def property_equals(self, model_object, key, value):
        current_value = model_object.__getattribute__(key)
        if current_value != value:
            self.add_failure(
                FailureType.ERROR,
                f"{model_object.get_model_header()} has '{key}' set to '{current_value}' but '{value}' is required.",
            )


class LivePatchType(Enum):
    ADD = 1
    REMOVE = 2
    CHANGE = 3


class LivePatchApplyFn(Protocol):
    async def __call__(self, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        ...


@dataclasses.dataclass(frozen=True)
class LivePatch:
    patch_type: LivePatchType
    expected_object: Optional[ModelObject]
    current_object: Optional[ModelObject]
    changes: Optional[dict[str, Change]]
    parent_object: Optional[ModelObject]
    forced_update: bool
    fn: LivePatchApplyFn

    @classmethod
    def of_addition(
        cls, expected_object: ModelObject, parent_object: Optional[ModelObject], fn: LivePatchApplyFn
    ) -> LivePatch:
        return LivePatch(LivePatchType.ADD, expected_object, None, None, parent_object, False, fn)

    @classmethod
    def of_deletion(
        cls, current_object: ModelObject, parent_object: Optional[ModelObject], fn: LivePatchApplyFn
    ) -> LivePatch:
        return LivePatch(LivePatchType.REMOVE, None, current_object, None, parent_object, False, fn)

    @classmethod
    def of_changes(
        cls,
        expected_object: ModelObject,
        current_object: ModelObject,
        changes: dict[str, Change],
        parent_object: Optional[ModelObject],
        forced_update: bool,
        fn: LivePatchApplyFn,
    ) -> LivePatch:
        return LivePatch(
            LivePatchType.CHANGE, expected_object, current_object, changes, parent_object, forced_update, fn
        )

    def requires_secrets(self) -> bool:
        match self.patch_type:
            case LivePatchType.ADD:
                assert self.expected_object is not None
                return self.expected_object.contains_secrets()

            case LivePatchType.REMOVE:
                return False

            case LivePatchType.CHANGE:
                assert self.expected_object is not None
                return self.expected_object.contains_secrets()

    async def apply(self, org_id: str, provider: GitHubProvider) -> None:
        await self.fn(self, org_id, provider)

    def __repr__(self) -> str:
        obj = self.expected_object if self.expected_object is not None else self.current_object
        assert obj is not None
        return f"{self.patch_type.name} - {obj.get_model_header(self.parent_object)}"


@dataclasses.dataclass
class PatchContext(object):
    org_id: str
    org_settings: ModelObject


@dataclasses.dataclass
class LivePatchContext(object):
    org_id: str
    update_webhooks: bool
    update_secrets: bool
    update_filter: str
    expected_org_settings: ModelObject
    modified_org_settings: dict[str, Change] = dataclasses.field(default_factory=dict)


class LivePatchHandler(Protocol):
    def __call__(self, patch: LivePatch) -> None:
        ...


@dataclasses.dataclass
class ModelObject(ABC):
    """
    The abstract base class for any model object.
    """

    def __post_init__(self):
        """
        Assigns to all field which are UNSET their default value, if one is available.
        """
        for field in self.all_fields():
            value = self.__getattribute__(field.name)
            if is_unset(value):
                if field.default is not dataclasses.MISSING:
                    self.__setattr__(field.name, field.default)
                elif field.default_factory is not dataclasses.MISSING:
                    self.__setattr__(field.name, field.default_factory())

    @property
    @abstractmethod
    def model_object_name(self) -> str:
        ...

    def is_keyed(self) -> bool:
        """Indicates whether the ModelObject is keyed by a property"""
        return any(field.metadata.get("key", False) for field in self.all_fields())

    def get_key(self) -> str:
        """Returns the key property of this ModelObject if it keyed"""
        assert self.is_keyed()
        return next(
            filter(
                lambda field: field.metadata.get("key", False) is True,
                self.all_fields(),
            )
        ).name

    def get_key_value(self) -> Any:
        """Returns the value of the key property"""
        return self.__getattribute__(self.get_key())

    def get_all_key_values(self) -> list[Any]:
        """Returns a list of all values by which this ModelObject is keyed"""
        return [self.get_key_value()]

    @abstractmethod
    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        ...

    # noinspection PyMethodMayBeStatic
    def execute_custom_validation_if_present(self, context: ValidationContext, filename: str) -> None:
        validate_script = os.path.join(context.template_dir, filename)
        if os.path.exists(validate_script):
            exec(open(validate_script).read())

    def get_difference_from(self, other: ModelObject) -> dict[str, Change[T]]:
        if not isinstance(other, self.__class__):
            raise ValueError(f"'types do not match: {type(self)}' != '{type(other)}'")

        diff_result: dict[str, Change[T]] = {}
        for key in self.keys(
            for_diff=True,
            for_patch=False,
            include_nested_models=False,
            exclude_unset_keys=True,
        ):
            to_value = self.__getattribute__(key)
            from_value = other.__getattribute__(key)

            if not other.is_key_valid_for_diff_computation(key):
                continue

            if is_unset(from_value):
                continue

            if is_different_ignoring_order(to_value, from_value):
                diff_result[key] = Change(from_value, to_value)

        return diff_result

    def get_patch_to(self, other: ModelObject) -> dict[str, Any]:
        if not isinstance(other, self.__class__):
            raise ValueError(f"'types do not match: {type(self)}' != '{type(other)}'")

        patch_result = {}
        for key in self.keys(
            for_diff=False,
            for_patch=True,
            include_nested_models=False,
            exclude_unset_keys=True,
        ):
            value = self.__getattribute__(key)
            other_value = other.__getattribute__(key)

            if is_unset(other_value):
                continue

            patch_needed, diff = patch_to_other(value, other_value)
            if patch_needed is True:
                patch_result[key] = diff

        return patch_result

    @classmethod
    def all_fields(cls) -> list[dataclasses.Field]:
        return [field for field in dataclasses.fields(cls)]

    @classmethod
    def model_fields(cls) -> list[dataclasses.Field]:
        return [field for field in dataclasses.fields(cls) if not cls.is_external_only(field)]

    @classmethod
    def model_only_fields(cls) -> list[dataclasses.Field]:
        return [field for field in dataclasses.fields(cls) if cls.is_model_only(field)]

    @classmethod
    def provider_fields(cls) -> list[dataclasses.Field]:
        return [
            field
            for field in dataclasses.fields(cls)
            if not cls.is_external_only(field)
            and not cls.is_model_only(field)
            and not cls.is_read_only(field)
            and not cls.is_nested_model(field)
        ]

    @classmethod
    def _get_field(cls, key: str) -> dataclasses.Field:
        for field in dataclasses.fields(cls):
            if field.name == key:
                return field

        raise ValueError(f"unknown key {key}")

    @staticmethod
    def is_external_only(field: dataclasses.Field) -> bool:
        return field.metadata.get("external_only", False) is True

    @staticmethod
    def is_read_only(field: dataclasses.Field) -> bool:
        return field.metadata.get("read_only", False) is True

    @classmethod
    def is_read_only_key(cls, key: str) -> bool:
        return cls.is_read_only(cls._get_field(key))

    @classmethod
    def is_nested_model_key(cls, key: str) -> bool:
        return cls.is_nested_model(cls._get_field(key))

    @staticmethod
    def is_model_only(field: dataclasses.Field) -> bool:
        return field.metadata.get("model_only", False) is True

    @staticmethod
    def is_nested_model(field: dataclasses.Field) -> bool:
        return field.metadata.get("nested_model", False) is True

    def get_model_objects(self) -> Iterator[tuple[ModelObject, ModelObject]]:
        yield from []

    def get_model_header(self, parent_object: Optional[ModelObject] = None) -> str:
        header = style(self.model_object_name, bright=True)

        if self.is_keyed():
            key = self.get_key()
            header = header + f'[{key}="{style(self.get_key_value(), bright=True)}"'

            if isinstance(parent_object, ModelObject):
                header = (
                    header
                    + f", {parent_object.model_object_name}="
                    + f'"{style(parent_object.get_key_value(), bright=True)}"'
                )

            header = header + "]"
        elif isinstance(parent_object, ModelObject) and parent_object.is_keyed():
            header = header + "["
            header = (
                header
                + f"{parent_object.model_object_name}="
                + f'"{style(parent_object.get_key_value(), bright=True)}"'
            )
            header = header + "]"

        return header

    @classmethod
    @abstractmethod
    def from_model_data(cls, data: dict[str, Any]):
        ...

    @classmethod
    @abstractmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]):
        ...

    @classmethod
    @abstractmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        ...

    async def to_provider_data(self, org_id: str, provider: GitHubProvider) -> dict[str, Any]:
        return await self.dict_to_provider_data(org_id, self.to_model_dict(), provider)

    @classmethod
    async def changes_to_provider(
        cls, org_id: str, data: dict[str, Change[Any]], provider: GitHubProvider
    ) -> dict[str, Any]:
        return await cls.dict_to_provider_data(org_id, {key: change.to_value for key, change in data.items()}, provider)

    @classmethod
    async def dict_to_provider_data(cls, org_id: str, data: dict[str, Any], provider: GitHubProvider) -> dict[str, Any]:
        mapping = await cls.get_mapping_to_provider(org_id, data, provider)
        return bend(mapping, data)

    @classmethod
    @abstractmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        ...

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        return True

    def is_key_valid_for_diff_computation(self, key: str) -> bool:
        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return self.include_field_for_diff_computation(field)

    def include_for_live_patch(self) -> bool:
        """
        Indicates if this ModelObject should be considered when generating a live patch.

        This method can be overridden if certain object should be ignored depending on their values,
        e.g. secrets that contain only a dummy secret
        """
        return True

    def include_existing_object_for_live_patch(self, org_id: str, parent_object: Optional[ModelObject]) -> bool:
        """
        Indicates if this live ModelObject should be considered when generating a live patch.

        This method can be overridden if certain live objects should be ignored in some cases, e.g. environments
        """
        return True

    def keys(
        self,
        for_diff: bool = False,
        for_patch: bool = False,
        include_nested_models: bool = False,
        exclude_unset_keys: bool = True,
    ) -> list[str]:
        result = list()

        for field in self.model_fields():
            if for_diff is True and not self.include_field_for_diff_computation(field):
                continue

            if for_patch is True and not self.include_field_for_patch_computation(field):
                continue

            if (for_diff or for_patch) and self.is_model_only(field):
                continue

            if include_nested_models is False and self.is_nested_model(field):
                continue

            if exclude_unset_keys:
                value = self.__getattribute__(field.name)
                if not is_unset(value):
                    result.append(field.name)
            else:
                result.append(field.name)

        return result

    def to_model_dict(self, for_diff: bool = False, include_nested_models: bool = False) -> dict[str, Any]:
        result = {}

        for key in self.keys(
            for_diff=for_diff,
            include_nested_models=include_nested_models,
            exclude_unset_keys=True,
        ):
            value = self.__getattribute__(key)
            if self.is_nested_model_key(key):
                result[key] = cast(ModelObject, value).to_model_dict(for_diff, include_nested_models)
            else:
                result[key] = value

        return result

    def contains_secrets(self) -> bool:
        return False

    def resolve_secrets(self, secret_resolver: Callable[[str], str]) -> None:
        pass

    def copy_secrets(self, other_object: ModelObject) -> None:
        pass

    @abstractmethod
    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        ...

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)

        template_function = self.get_jsonnet_template_function(jsonnet_config, False)
        assert template_function is not None

        if self.is_keyed():
            key = patch.pop(self.get_key())
            printer.print(f"{template_function}('{key}')")
        else:
            printer.print(f"{template_function}")

        write_patch_object_as_json(patch, printer)

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: Optional[ModelObject],
        current_object: Optional[ModelObject],
        parent_object: Optional[ModelObject],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        if current_object is None:
            assert isinstance(expected_object, cls)
            handler(LivePatch.of_addition(expected_object, parent_object, expected_object.apply_live_patch))
            return

        if expected_object is None:
            assert isinstance(current_object, cls)
            handler(LivePatch.of_deletion(current_object, parent_object, current_object.apply_live_patch))
            return

        assert isinstance(expected_object, cls)
        assert isinstance(current_object, cls)

        modified_rule: dict[str, Change[Any]] = expected_object.get_difference_from(current_object)

        if len(modified_rule) > 0:
            handler(
                LivePatch.of_changes(
                    expected_object,
                    current_object,
                    modified_rule,
                    parent_object,
                    False,
                    expected_object.apply_live_patch,
                )
            )

    @classmethod
    def generate_live_patch_of_list(
        cls,
        expected_objects: Sequence[MT],
        current_objects: Sequence[MT],
        parent_object: Optional[MT],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        expected_objects_by_key = associate_by_key(expected_objects, lambda x: x.get_key_value())
        expected_objects_by_all_keys = multi_associate_by_key(expected_objects, lambda x: x.get_all_key_values())

        for current_object in current_objects:
            key = current_object.get_key_value()

            expected_object = expected_objects_by_all_keys.get(key)
            if expected_object is None:
                if current_object.include_existing_object_for_live_patch(context.org_id, parent_object):
                    cls.generate_live_patch(None, current_object, parent_object, context, handler)
                continue

            if expected_object.include_for_live_patch():
                cls.generate_live_patch(expected_object, current_object, parent_object, context, handler)

            for k in expected_object.get_all_key_values():
                expected_objects_by_all_keys.pop(k)
            expected_objects_by_key.pop(expected_object.get_key_value())

        for _, expected_object in expected_objects_by_key.items():
            if expected_object.include_for_live_patch():
                cls.generate_live_patch(expected_object, None, parent_object, context, handler)

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        ...



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from abc import ABC, abstractmethod
from typing import Any, Optional

from otterdog.config import OrganizationConfig, OtterdogConfig
from otterdog.utils import Change, IndentingPrinter, is_unset, style


class Operation(ABC):
    _DEFAULT_WIDTH: int = 33

    def __init__(self) -> None:
        self._config: Optional[OtterdogConfig] = None
        self._printer: Optional[IndentingPrinter] = None

    def init(self, config: OtterdogConfig, printer: IndentingPrinter) -> None:
        self._config = config
        self._printer = printer

    @property
    def config(self) -> OtterdogConfig:
        assert self._config is not None
        return self._config

    @property
    def printer(self) -> IndentingPrinter:
        assert self._printer is not None
        return self._printer

    @printer.setter
    def printer(self, value: IndentingPrinter):
        self._printer = value

    @abstractmethod
    def pre_execute(self) -> None:
        ...

    @abstractmethod
    async def execute(self, org_config: OrganizationConfig) -> int:
        ...

    def post_execute(self) -> None:
        pass

    def print_dict(
        self,
        data: dict[str, Any],
        item_header: str,
        action: str,
        color: str,
        key_value_separator: str = "=",
        value_separator: str = "",
    ) -> None:
        prefix = f"{style(action, fg=color)} " if action else ""
        closing_prefix = prefix

        if item_header:
            self.printer.print(f"{prefix}{item_header} ")
        self._print_dict_internal(data, prefix, closing_prefix, False, key_value_separator, value_separator)

    def _print_internal(
        self,
        data: Any,
        prefix: str,
        closing_prefix: str,
        include_prefix: bool,
        key_value_separator: str,
        value_separator: str,
    ):
        if isinstance(data, dict):
            self._print_dict_internal(
                data, prefix, closing_prefix, include_prefix, key_value_separator, value_separator
            )
        elif isinstance(data, list):
            self._print_list_internal(data, prefix, closing_prefix, key_value_separator, value_separator)
        else:
            self.printer.println(f"{prefix if include_prefix else ''}{self._get_value(data)}{value_separator}")

    def _print_dict_internal(
        self,
        data: dict[str, Any],
        prefix: str,
        closing_prefix: str,
        include_prefix: bool,
        key_value_separator: str,
        value_separator: str,
    ):
        self.printer.println(f"{prefix if include_prefix else ''}{{")
        self.printer.level_up()
        for key, value in sorted(data.items()):
            self.printer.print(f"{prefix}{key.ljust(self._DEFAULT_WIDTH, ' ')} {key_value_separator} ")
            self._print_internal(value, prefix, closing_prefix, False, key_value_separator, value_separator)

        self.printer.level_down()
        self.printer.println(f"{closing_prefix}}}{value_separator}")

    def _print_list_internal(
        self, data: list[Any], prefix: str, closing_prefix: str, key_value_separator: str, value_separator: str
    ):
        if len(data) > 0:
            self.printer.println("[")
            self.printer.level_up()
            for entry in data:
                self._print_internal(entry, prefix, closing_prefix, True, key_value_separator, value_separator)

            self.printer.level_down()
            self.printer.println(f"{closing_prefix}],")
        else:
            self.printer.println(f"[]{value_separator}")

    @staticmethod
    def _get_value(value: Any) -> str:
        if value is None:
            return "null"
        if isinstance(value, bool):
            return str(value).lower()
        else:
            return f'"{value}"'

    def print_modified_dict(
        self,
        data: dict[str, Change[Any]],
        item_header: str,
        redacted_keys: Optional[set[str]] = None,
        forced_update: bool = False,
    ) -> None:
        prefix = style("! ", fg="magenta") if forced_update else style("~ ", fg="yellow")
        closing_prefix = prefix
        color = "magenta" if forced_update else "yellow"

        self.printer.println(f"\n{prefix}{item_header} {{")
        self.printer.level_up()

        for key, change in sorted(data.items()):
            current_value = change.from_value
            expected_value = change.to_value

            if isinstance(expected_value, dict):
                self.printer.println(f"{prefix}{key.ljust(self._DEFAULT_WIDTH, ' ')} = {{")
                self.printer.level_up()

                processed_keys = set()
                for k, v in sorted(expected_value.items()):
                    c_v = current_value.get(k) if current_value is not None else None

                    if v != c_v:
                        self.printer.println(
                            f"{prefix}{k.ljust(self._DEFAULT_WIDTH, ' ')} ="
                            f' {self._get_value(c_v)} {style("->", fg=color)} {self._get_value(v)}'
                        )

                    processed_keys.add(k)

                if current_value is not None:
                    for k, v in sorted(current_value.items()):
                        if k not in processed_keys:
                            self.printer.println(
                                f"{style('- ', fg='red')}{k.ljust(self._DEFAULT_WIDTH, ' ')} =" f" {self._get_value(v)}"
                            )

                self.printer.println("}")
                self.printer.level_down()
            else:

                def should_redact(value: Any) -> bool:
                    if is_unset(value) or value is None:
                        return False

                    if redacted_keys is not None and key is not None and key in redacted_keys:
                        return True
                    else:
                        return False

                c_v = "<redacted>" if should_redact(current_value) else current_value
                e_v = "<redacted>" if should_redact(expected_value) else expected_value

                self.printer.println(
                    f"{prefix}{key.ljust(self._DEFAULT_WIDTH, ' ')} ="
                    f' {self._get_value(c_v)} {style("->", fg=color)} {self._get_value(e_v)}'
                )

        self.printer.level_down()
        self.printer.println(f"{closing_prefix}}}")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

"""Data classes for events received via webhook from GitHub"""


from __future__ import annotations

from abc import ABC
from typing import Optional

from pydantic import BaseModel


class Installation(BaseModel):
    """The installation that is associated with the event."""

    id: int
    node_id: str


class Organization(BaseModel):
    """The organization that is associated with the event."""

    login: str
    id: int
    node_id: str


class Repository(BaseModel):
    """A reference to the repository."""

    id: int
    node_id: str
    name: str
    full_name: str
    private: bool
    owner: Actor
    default_branch: str


class Actor(BaseModel):
    """An actor, can be either of type 'User' or 'Organization'."""

    login: str
    id: int
    node_id: str
    type: str


class Ref(BaseModel):
    """A ref in a repository."""

    label: str
    ref: str
    sha: str
    user: Actor
    repo: Repository


class PullRequest(BaseModel):
    """Represents a pull request."""

    id: int
    node_id: str
    number: int
    state: str
    locked: bool
    title: str
    body: Optional[str] = None
    draft: bool
    merged: bool
    merge_commit_sha: Optional[str] = None
    user: Actor

    head: Ref
    base: Ref


class Comment(BaseModel):
    """Represents a comment in an issue."""

    id: int
    node_id: str
    user: Actor
    body: str
    created_at: str
    updated_at: str


class Issue(BaseModel):
    """Represents an issue"""

    number: int
    node_id: str
    title: str
    state: str
    draft: bool
    body: Optional[str]
    html_url: str


class Event(ABC, BaseModel):
    """Base class of events"""

    installation: Optional[Installation] = None
    organization: Optional[Organization] = None
    sender: Actor


class PullRequestEvent(Event):
    """A payload sent for pull request specific events."""

    action: str
    number: int
    pull_request: PullRequest
    repository: Repository


class PushEvent(Event):
    """A payload sent for push events."""

    ref: str
    before: str
    after: str

    repository: Repository

    created: bool
    deleted: bool
    forced: bool


class IssueCommentEvent(Event):
    """A payload sent for issue comment events."""

    action: str
    issue: Issue
    comment: Comment
    repository: Repository



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Any

from otterdog.config import OrganizationConfig
from otterdog.providers.github import GitHubProvider
from otterdog.utils import is_info_enabled, style

from . import Operation


class ListAppsOperation(Operation):
    """
    Lists installed GitHub apps for an organization.
    """

    def __init__(self, json_output: bool):
        super().__init__()
        self.all_apps: dict[str, Any] = {}
        self.json_output = json_output

    def pre_execute(self) -> None:
        if not self.json_output or is_info_enabled():
            self.printer.println("Listing app installations:")

    def post_execute(self) -> None:
        if self.json_output is True:
            if is_info_enabled():
                self.printer.println()
            apps = [v for k, v in sorted(self.all_apps.items())]
            self.printer.println(json.dumps(apps, indent=2))

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id

        if not self.json_output or is_info_enabled():
            self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
            self.printer.level_up()

        try:
            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                apps = await provider.rest_api.org.get_app_installations(github_id)

            if not self.json_output:
                for app in apps:
                    filtered = {key: app[key] for key in ["app_id", "permissions"]}
                    self.printer.println()
                    self.print_dict(filtered, f"app['{app['app_slug']}']", "", "black")
            else:
                if is_info_enabled():
                    self.printer.println(f"Found {len(apps)} app installations.")

                for app in apps:
                    app_slug = str(app["app_slug"])
                    filtered = {key: app[key] for key in ["app_id", "app_slug", "permissions"]}
                    self.all_apps[app_slug] = filtered

            return 0
        finally:
            if not self.json_output or is_info_enabled():
                self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from otterdog.config import OrganizationConfig
from otterdog.providers.github import GitHubProvider
from otterdog.utils import get_approval, style

from . import Operation


class FetchOperation(Operation):
    """
    Fetches the current configuration from the meta-data repository of an organization.
    """

    def __init__(self, force_processing: bool, pull_request: str):
        super().__init__()
        self.force_processing = force_processing
        self.pull_request = pull_request

    def pre_execute(self) -> None:
        self.printer.println("Fetching organization configurations:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")

        org_file_name = jsonnet_config.org_config_file

        if os.path.exists(org_file_name) and not self.force_processing:
            self.printer.println()
            self.printer.println(style("Definition already exists", bright=True) + f" at '{org_file_name}'.")
            self.printer.println("  Performing this action will overwrite its contents.")
            self.printer.println("  Do you want to continue? (Only 'yes' or 'y' will be accepted to approve)\n")

            self.printer.print(f"{style('Enter a value:', bright=True)} ")
            if not get_approval():
                self.printer.println("\nFetch cancelled.")
                return 1

        self.printer.level_up()

        try:
            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                try:
                    if self.pull_request is not None:
                        ref = await provider.get_ref_for_pull_request(
                            org_config.github_id, org_config.config_repo, self.pull_request
                        )
                    else:
                        ref = None

                    definition = await provider.get_content(
                        org_config.github_id,
                        org_config.config_repo,
                        f"otterdog/{github_id}.jsonnet",
                        ref,
                    )
                except RuntimeError:
                    self.printer.print_error(f"failed to fetch definition from repo '{org_config.config_repo}'")
                    return 1

            output_dir = jsonnet_config.org_dir
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)

            with open(org_file_name, "w") as file:
                file.write(definition)

            if ref is not None:
                self.printer.println(
                    f"organization definition fetched from pull request " f"#{self.pull_request} to '{org_file_name}'"
                )
            else:
                self.printer.println(f"organization definition fetched from default branch to '{org_file_name}'")

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from otterdog.config import OrganizationConfig
from otterdog.models import FailureType
from otterdog.models.github_organization import GitHubOrganization
from otterdog.utils import is_info_enabled, style

from . import Operation


class ValidateOperation(Operation):
    """
    Validates local organization configurations.
    """

    def __init__(self):
        super().__init__()

    def pre_execute(self) -> None:
        self.printer.println("Validating organization configurations:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
        self.printer.level_up()

        try:
            org_file_name = jsonnet_config.org_config_file

            if not os.path.exists(org_file_name):
                self.printer.print_error(
                    f"configuration file '{org_file_name}' does not yet exist, run 'fetch-config' or 'import' first."
                )
                return 1

            try:
                organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
            except RuntimeError as ex:
                self.printer.print_error(f"Validation failed\nfailed to load configuration: {str(ex)}")
                return 1

            validation_infos, validation_warnings, validation_errors = self.validate(
                organization, jsonnet_config.template_dir
            )
            validation_count = validation_infos + validation_warnings + validation_errors

            if validation_count == 0:
                self.printer.println(style("Validation succeeded", fg="green"))
            else:
                if validation_errors == 0:
                    self.printer.println(
                        f"{style('Validation succeeded', fg='green')}: "
                        f"{validation_infos} info(s), {validation_warnings} warning(s), "
                        f"{validation_errors} error(s)"
                    )
                else:
                    self.printer.println(
                        f"{style('Validation failed', fg='red')}: "
                        f"{validation_infos} info(s), {validation_warnings} warning(s), "
                        f"{validation_errors} error(s)"
                    )

            if validation_infos > 0 and not is_info_enabled():
                self.printer.level_up()
                self.printer.println(
                    "in order to print validation infos, enable printing info messages by adding '-v' flag."
                )
                self.printer.level_down()

            return validation_errors
        finally:
            self.printer.level_down()

    def validate(self, organization: GitHubOrganization, template_dir: str) -> tuple[int, int, int]:
        if organization.secrets_resolved is True:
            raise RuntimeError("validation requires an unresolved model.")

        context = organization.validate(template_dir)

        validation_infos = 0
        validation_warnings = 0
        validation_errors = 0

        for failure_type, message in context.validation_failures:
            match failure_type:
                case FailureType.INFO:
                    self.printer.print_info(message)
                    validation_infos += 1

                case FailureType.WARNING:
                    self.printer.print_warn(message)
                    validation_warnings += 1

                case FailureType.ERROR:
                    self.printer.print_error(message)
                    validation_errors += 1

        return validation_infos, validation_warnings, validation_errors



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
from typing import Optional

from otterdog.config import OrganizationConfig
from otterdog.jsonnet import JsonnetConfig
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider

from .plan import PlanOperation


class LocalPlanOperation(PlanOperation):
    def __init__(self, suffix: str, update_webhooks: bool, update_secrets: bool, update_filter: str) -> None:
        super().__init__(
            no_web_ui=False, update_webhooks=update_webhooks, update_secrets=update_secrets, update_filter=update_filter
        )

        self.suffix = suffix
        self._other_org: Optional[GitHubOrganization] = None

    @property
    def other_org(self) -> GitHubOrganization:
        assert self._other_org is not None
        return self._other_org

    def pre_execute(self) -> None:
        self.printer.println("Printing local diff:")
        self.print_legend()

    def verbose_output(self):
        return False

    def resolve_secrets(self) -> bool:
        return False

    def setup_github_client(self, org_config: OrganizationConfig) -> GitHubProvider:
        return GitHubProvider(None)

    async def load_current_org(self, github_id: str, jsonnet_config: JsonnetConfig) -> GitHubOrganization:
        other_org_file_name = jsonnet_config.org_config_file + self.suffix

        if not os.path.exists(other_org_file_name):
            raise RuntimeError(f"configuration file '{other_org_file_name}' does not exist")

        return GitHubOrganization.load_from_file(github_id, other_org_file_name, self.config)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Any, Optional

from aiohttp_client_cache.backends import FileBackend
from aiohttp_client_cache.session import CachedSession as AsyncCachedSession
from requests import Response
from requests_cache import CachedSession

from otterdog.providers.github.auth import AuthStrategy
from otterdog.providers.github.exception import BadCredentialsException, GitHubException
from otterdog.utils import is_debug_enabled, is_trace_enabled, print_debug, print_trace

_AIOHTTP_CACHE_DIR = ".cache/async_http"
_REQUESTS_CACHE_DIR = ".cache/http"


class Requester:
    def __init__(self, auth_strategy: AuthStrategy, base_url: str, api_version: str):
        self._base_url = base_url
        self._auth = auth_strategy.get_auth()

        self._headers = {
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": api_version,
            "X-Github-Next-Global-ID": "1",
        }

        # enable logging for requests_cache
        # import logging
        #
        #
        # logging.basicConfig(level='DEBUG')

        self._session: CachedSession = CachedSession(
            _REQUESTS_CACHE_DIR,
            backend="filesystem",
            use_cache_dir=False,
            cache_control=True,
            allowable_methods=["GET"],
        )
        self._session.auth = self._auth

    def close(self) -> None:
        self._session.close()

    def _build_url(self, url_path: str) -> str:
        return f"{self._base_url}{url_path}"

    def request_paged_json(
        self,
        method: str,
        url_path: str,
        data: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, str]] = None,
    ) -> list[dict[str, Any]]:
        result = []
        current_page = 1
        while current_page > 0:
            query_params = {"per_page": "100", "page": current_page}
            if params is not None:
                query_params.update(params)

            response: list[dict[str, Any]] = self.request_json(method, url_path, data, query_params)

            if len(response) == 0:
                current_page = -1
            else:
                for item in response:
                    result.append(item)

                current_page += 1

        return result

    async def async_request_paged_json(
        self,
        method: str,
        url_path: str,
        data: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, str]] = None,
    ) -> list[dict[str, Any]]:
        result = []
        current_page = 1
        while current_page > 0:
            query_params = {"per_page": "100", "page": current_page}
            if params is not None:
                query_params.update(params)

            response: list[dict[str, Any]] = await self.async_request_json(method, url_path, data, query_params)

            if len(response) == 0:
                current_page = -1
            else:
                for item in response:
                    result.append(item)

                current_page += 1

        return result

    def request_json(
        self,
        method: str,
        url_path: str,
        data: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Any:
        input_data = None
        if data is not None:
            input_data = json.dumps(data)

        response = self.request_raw(method, url_path, input_data, params)
        self._check_response(response.url, response.status_code, response.text)
        return response.json()

    async def async_request_json(
        self,
        method: str,
        url_path: str,
        data: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Any:
        input_data = None
        if data is not None:
            input_data = json.dumps(data)

        status, body = await self.async_request_raw(method, url_path, input_data, params)
        self._check_response(url_path, status, body)
        return json.loads(body)

    def request_raw(
        self,
        method: str,
        url_path: str,
        data: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
        stream: bool = False,
        force_refresh: bool = False,
    ) -> Response:
        print_trace(f"'{method}' url = {url_path}, data = {data}, headers = {self._headers}")

        response = self._session.request(
            method,
            url=self._build_url(url_path),
            headers=self._headers,
            refresh=True,
            force_refresh=force_refresh,
            params=params,
            data=data,
            stream=stream,
        )

        if is_debug_enabled():
            print_debug(f"'{method}' {url_path}: rate-limit-used = {response.headers.get('x-ratelimit-used', None)}")

        if is_trace_enabled():
            print_trace(f"'{method}' result = ({response.status_code}, {response.text})")

        return response

    async def async_request_raw(
        self,
        method: str,
        url_path: str,
        data: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> tuple[int, str]:
        print_trace(f"async '{method}' url = {url_path}, data = {data}, headers = {self._headers}")

        headers = self._headers.copy()
        self._auth.update_headers_with_authorization(headers)

        async with AsyncCachedSession(cache=FileBackend(cache_name=_AIOHTTP_CACHE_DIR, use_temp=False)) as session:
            url = self._build_url(url_path)
            async with session.request(
                method, url=url, headers=headers, params=params, data=data, refresh=True
            ) as response:
                text = await response.text()
                status = response.status

                if is_debug_enabled():
                    if not response.from_cache:  # type: ignore
                        print_debug(
                            f"'{method}' {url_path}: rate-limit-used = {response.headers.get('x-ratelimit-used', None)}"
                        )

                if is_trace_enabled():
                    print_trace(f"async '{method}' result = ({status}, {text})")

                return status, text

    def _check_response(self, url_path: str, status_code: int, body: str) -> None:
        if status_code >= 400:
            self._create_exception(self._build_url(url_path), status_code, body)

    @staticmethod
    def _create_exception(url: str, status_code: int, body: str):
        if status_code == 401:
            raise BadCredentialsException(url, status_code, body)
        else:
            raise GitHubException(url, status_code, body)



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Optional

from otterdog.utils import print_debug

from . import RestApi, RestClient


class CommitClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def create_commit_status(
        self,
        org_id: str,
        repo_name: str,
        sha: str,
        state: str,
        context: str,
        description: Optional[str] = None,
        target_url: Optional[str] = None,
    ) -> None:
        print_debug(f"creating a commit status for sha '{sha}'")

        data = {"state": state, "target_url": target_url, "description": description, "context": context}
        status, body = await self.requester.async_request_raw(
            "POST", f"/repos/{org_id}/{repo_name}/statuses/{sha}", data=json.dumps(data)
        )

        if status != 201:
            raise RuntimeError(f"failed creating commit status for '{org_id}/{repo_name}/{sha}'\n{status}: {body}")

        print_debug(f"created commit status for sha '{org_id}/{repo_name}/{sha}'")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
import re
from typing import Any

from otterdog.providers.github.exception import GitHubException
from otterdog.utils import print_debug, print_trace, print_warn

from . import RestApi, RestClient, encrypt_value


class OrgClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_settings(self, org_id: str, included_keys: set[str]) -> dict[str, Any]:
        print_debug(f"retrieving settings for org '{org_id}'")

        try:
            settings = await self.requester.async_request_json("GET", f"/orgs/{org_id}")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving settings for organization '{org_id}':\n{ex}").with_traceback(tb)

        if "security_managers" in included_keys:
            security_managers = await self.list_security_managers(org_id)
            settings["security_managers"] = security_managers

        result = {}
        for k, v in settings.items():
            if k in included_keys:
                result[k] = v
                print_trace(f"retrieved setting for '{k}' = '{v}'")

        return result

    async def update_settings(self, org_id: str, data: dict[str, Any]) -> None:
        print_debug(f"updating settings for organization '{org_id}'")

        try:
            await self.requester.async_request_json("PATCH", f"/orgs/{org_id}", data)
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to update settings for organization '{org_id}':\n{ex}").with_traceback(tb)

        if "security_managers" in data:
            await self.update_security_managers(org_id, data["security_managers"])

        print_debug(f"updated {len(data)} setting(s)")

    async def list_security_managers(self, org_id: str) -> list[str]:
        print_debug(f"retrieving security managers for organization {org_id}")

        try:
            result = await self.requester.async_request_json("GET", f"/orgs/{org_id}/security-managers")
            return list(map(lambda x: x["slug"], result))
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(
                f"failed retrieving security managers for organization " f"'{org_id}':\n{ex}"
            ).with_traceback(tb)

    async def update_security_managers(self, org_id: str, security_managers: list[str]) -> None:
        print_debug(f"updating security managers for organization {org_id}")

        current_managers = set(await self.list_security_managers(org_id))

        # first, add all security managers that are not yet configured.
        for team_slug in security_managers:
            if team_slug in current_managers:
                current_managers.remove(team_slug)
            else:
                await self.add_security_manager_team(org_id, team_slug)

        # second, remove the current managers that are left.
        for team_slug in current_managers:
            await self.remove_security_manager_team(org_id, team_slug)

    async def add_security_manager_team(self, org_id: str, team_slug: str) -> None:
        print_debug(f"adding team {team_slug} to security managers for organization {org_id}")

        status, body = await self.requester.async_request_raw(
            "PUT", f"/orgs/{org_id}/security-managers/teams/{team_slug}"
        )

        if status == 204:
            print_debug(f"added team {team_slug} to security managers for organization {org_id}")
        elif status == 404:
            print_warn(
                f"failed to add team '{team_slug}' to security managers for organization {org_id}: " f"team not found"
            )
        else:
            raise RuntimeError(
                f"failed adding team '{team_slug}' to security managers of organization '{org_id}'"
                f"\n{status}: {body}"
            )

    async def remove_security_manager_team(self, org_id: str, team_slug: str) -> None:
        print_debug(f"removing team {team_slug} from security managers for organization {org_id}")

        status, body = await self.requester.async_request_raw(
            "DELETE", f"/orgs/{org_id}/security-managers/teams/{team_slug}"
        )
        if status != 204:
            raise RuntimeError(
                f"failed removing team '{team_slug}' from security managers of organization '{org_id}'"
                f"\n{status}: {body}"
            )

        print_debug(f"removed team {team_slug} from security managers for organization {org_id}")

    async def get_webhooks(self, org_id: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving org webhooks for org '{org_id}'")

        try:
            return await self.requester.async_request_json("GET", f"/orgs/{org_id}/hooks")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving webhooks for org '{org_id}':\n{ex}").with_traceback(tb)

    async def update_webhook(self, org_id: str, webhook_id: int, webhook: dict[str, Any]) -> None:
        print_debug(f"updating org webhook '{webhook_id}' for organization {org_id}")

        try:
            await self.requester.async_request_json("PATCH", f"/orgs/{org_id}/hooks/{webhook_id}", webhook)
            print_debug(f"updated webhook {webhook_id}")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to update org webhook {webhook_id}:\n{ex}").with_traceback(tb)

    async def add_webhook(self, org_id: str, data: dict[str, Any]) -> None:
        url = data["config"]["url"]
        print_debug(f"adding org webhook with url '{url}'")

        # mandatory field "name" = "web"
        data["name"] = "web"

        try:
            await self.requester.async_request_json("POST", f"/orgs/{org_id}/hooks", data)
            print_debug(f"added org webhook with url '{url}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to add org webhook with url '{url}':\n{ex}").with_traceback(tb)

    async def delete_webhook(self, org_id: str, webhook_id: int, url: str) -> None:
        print_debug(f"deleting org webhook with url '{url}'")

        status, _ = await self.requester.async_request_raw("DELETE", f"/orgs/{org_id}/hooks/{webhook_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete org webhook with url '{url}'")

        print_debug(f"removed org webhook with url '{url}'")

    async def get_repos(self, org_id: str) -> list[str]:
        print_debug(f"retrieving repos for organization {org_id}")

        params = {"type": "all"}
        try:
            repos = await self.requester.async_request_paged_json("GET", f"/orgs/{org_id}/repos", params)
            return [repo["name"] for repo in repos]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to retrieve repos for organization '{org_id}':\n{ex}").with_traceback(tb)

    async def get_secrets(self, org_id: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving secrets for org '{org_id}'")

        try:
            response = await self.requester.async_request_json("GET", f"/orgs/{org_id}/actions/secrets")

            secrets = response["secrets"]
            for secret in secrets:
                if secret["visibility"] == "selected":
                    secret["selected_repositories"] = await self._get_selected_repositories_for_secret(
                        org_id, secret["name"]
                    )
            return secrets
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed getting secrets for org '{org_id}':\n{ex}").with_traceback(tb)

    async def _get_selected_repositories_for_secret(self, org_id: str, secret_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving selected repositories for secret '{secret_name}'")

        try:
            url = f"/orgs/{org_id}/actions/secrets/{secret_name}/repositories"
            response = await self.requester.async_request_json("GET", url)
            return response["repositories"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}").with_traceback(tb)

    async def _set_selected_repositories_for_secret(
        self, org_id: str, secret_name: str, selected_repository_ids: list[str]
    ) -> None:
        print_debug(f"setting selected repositories for secret '{secret_name}'")

        try:
            data = {"selected_repository_ids": selected_repository_ids}

            url = f"/orgs/{org_id}/actions/secrets/{secret_name}/repositories"
            status, _ = await self.requester.async_request_raw("PUT", url, json.dumps(data))
            if status != 204:
                raise RuntimeError(f"failed to update selected repositories for secret '{secret_name}'")
            else:
                print_debug(f"updated selected repositories for secret '{secret_name}'")

        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}").with_traceback(tb)

    async def update_secret(self, org_id: str, secret_name: str, secret: dict[str, Any]) -> None:
        print_debug(f"updating org secret '{secret_name}'")

        if "name" in secret:
            secret.pop("name")

        if "visibility" in secret:
            visibility = secret["visibility"]
        else:
            visibility = None

        if "selected_repository_ids" in secret:
            selected_repository_ids = secret.pop("selected_repository_ids")
        else:
            selected_repository_ids = None

        await self._encrypt_secret_inplace(org_id, secret)

        status, _ = await self.requester.async_request_raw(
            "PUT", f"/orgs/{org_id}/actions/secrets/{secret_name}", json.dumps(secret)
        )

        if status != 204:
            raise RuntimeError(f"failed to update org secret '{secret_name}'")

        if selected_repository_ids is not None and (visibility is None or visibility == "selected"):
            await self._set_selected_repositories_for_secret(org_id, secret_name, selected_repository_ids)

        print_debug(f"updated org secret '{secret_name}'")

    async def add_secret(self, org_id: str, data: dict[str, str]) -> None:
        secret_name = data.pop("name")
        print_debug(f"adding org secret '{secret_name}'")

        await self._encrypt_secret_inplace(org_id, data)

        status, _ = await self.requester.async_request_raw(
            "PUT", f"/orgs/{org_id}/actions/secrets/{secret_name}", json.dumps(data)
        )

        if status != 201:
            raise RuntimeError(f"failed to add org secret '{secret_name}'")

        print_debug(f"added org secret '{secret_name}'")

    async def _encrypt_secret_inplace(self, org_id: str, data: dict[str, Any]) -> None:
        if "value" in data:
            value = data.pop("value")
            key_id, public_key = await self.get_public_key(org_id)
            data["encrypted_value"] = encrypt_value(public_key, value)
            data["key_id"] = key_id

    async def delete_secret(self, org_id: str, secret_name: str) -> None:
        print_debug(f"deleting org secret '{secret_name}'")

        status, _ = await self.requester.async_request_raw("DELETE", f"/orgs/{org_id}/actions/secrets/{secret_name}")
        if status != 204:
            raise RuntimeError(f"failed to delete org secret '{secret_name}'")

        print_debug(f"removed org secret '{secret_name}'")

    async def get_variables(self, org_id: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving variables for org '{org_id}'")

        try:
            response = await self.requester.async_request_json("GET", f"/orgs/{org_id}/actions/variables")

            secrets = response["variables"]
            for secret in secrets:
                if secret["visibility"] == "selected":
                    secret["selected_repositories"] = await self._get_selected_repositories_for_variable(
                        org_id, secret["name"]
                    )
            return secrets
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed getting variables for org '{org_id}':\n{ex}").with_traceback(tb)

    async def _get_selected_repositories_for_variable(self, org_id: str, variable_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving selected repositories for variable '{variable_name}'")

        try:
            url = f"/orgs/{org_id}/actions/variables/{variable_name}/repositories"
            response = await self.requester.async_request_json("GET", url)
            return response["repositories"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}").with_traceback(tb)

    async def _set_selected_repositories_for_variable(
        self, org_id: str, variable_name: str, selected_repository_ids: list[str]
    ) -> None:
        print_debug(f"setting selected repositories for variable '{variable_name}'")

        try:
            data = {"selected_repository_ids": selected_repository_ids}

            url = f"/orgs/{org_id}/actions/variables/{variable_name}/repositories"
            status, _ = await self.requester.async_request_raw("PUT", url, json.dumps(data))
            if status != 204:
                raise RuntimeError(f"failed to update selected repositories for variable '{variable_name}'")

            print_debug(f"updated selected repositories for variable '{variable_name}'")

        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}").with_traceback(tb)

    async def update_variable(self, org_id: str, variable_name: str, variable: dict[str, Any]) -> None:
        print_debug(f"updating org variable '{variable_name}'")

        if "name" in variable:
            variable.pop("name")

        if "visibility" in variable:
            visibility = variable["visibility"]
        else:
            visibility = None

        if "selected_repository_ids" in variable:
            selected_repository_ids = variable.pop("selected_repository_ids")
        else:
            selected_repository_ids = None

        status, body = await self.requester.async_request_raw(
            "PATCH", f"/orgs/{org_id}/actions/variables/{variable_name}", json.dumps(variable)
        )
        if status != 204:
            raise RuntimeError(f"failed to update org variable '{variable_name}': {body}")

        if selected_repository_ids is not None and (visibility is None or visibility == "selected"):
            await self._set_selected_repositories_for_variable(org_id, variable_name, selected_repository_ids)

        print_debug(f"updated org variable '{variable_name}'")

    async def add_variable(self, org_id: str, data: dict[str, str]) -> None:
        variable_name = data.get("name")
        print_debug(f"adding org variable '{variable_name}'")

        status, body = await self.requester.async_request_raw(
            "POST", f"/orgs/{org_id}/actions/variables", json.dumps(data)
        )

        if status != 201:
            raise RuntimeError(f"failed to add org variable '{variable_name}': {body}")

        print_debug(f"added org variable '{variable_name}'")

    async def delete_variable(self, org_id: str, variable_name: str) -> None:
        print_debug(f"deleting org variable '{variable_name}'")

        status, body = await self.requester.async_request_raw(
            "DELETE", f"/orgs/{org_id}/actions/variables/{variable_name}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete org variable '{variable_name}': {body}")

        print_debug(f"removed org variable '{variable_name}'")

    async def get_public_key(self, org_id: str) -> tuple[str, str]:
        print_debug(f"retrieving org public key for org '{org_id}'")

        try:
            response = await self.requester.async_request_json("GET", f"/orgs/{org_id}/actions/secrets/public-key")
            return response["key_id"], response["key"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving org public key:\n{ex}").with_traceback(tb)

    async def get_team_ids(self, combined_slug: str) -> tuple[int, str]:
        print_debug("retrieving team ids")
        org_id, team_slug = re.split("/", combined_slug)

        try:
            response = await self.requester.async_request_json("GET", f"/orgs/{org_id}/teams/{team_slug}")
            return response["id"], response["node_id"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving team node id:\n{ex}").with_traceback(tb)

    async def get_teams(self, org_id: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving teams for org '{org_id}'")

        try:
            return await self.requester.async_request_json("GET", f"/orgs/{org_id}/teams")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving teams for org '{org_id}':\n{ex}").with_traceback(tb)

    async def get_app_installations(self, org_id: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving app installations for org '{org_id}'")

        try:
            response = await self.requester.async_request_json("GET", f"/orgs/{org_id}/installations")
            return response["installations"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed getting app installations for org '{org_id}':\n{ex}").with_traceback(tb)

    async def get_workflow_settings(self, org_id: str) -> dict[str, Any]:
        print_debug(f"retrieving workflow settings for org '{org_id}'")

        workflow_settings: dict[str, Any] = {}

        try:
            permissions = await self.requester.async_request_json("GET", f"/orgs/{org_id}/actions/permissions")
            workflow_settings.update(permissions)
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving workflow settings for org '{org_id}':\n{ex}").with_traceback(tb)

        if permissions["enabled_repositories"] == "selected":
            workflow_settings["selected_repositories"] = await self._get_selected_repositories_for_workflow_settings(
                org_id
            )
        else:
            workflow_settings["selected_repositories"] = None

        allowed_actions = permissions.get("allowed_actions", "none")
        if allowed_actions == "selected":
            workflow_settings.update(await self._get_selected_actions_for_workflow_settings(org_id))

        if allowed_actions != "none":
            workflow_settings.update(await self._get_default_workflow_permissions(org_id))

        return workflow_settings

    async def update_workflow_settings(self, org_id: str, data: dict[str, Any]) -> None:
        print_debug(f"updating workflow settings for org '{org_id}'")

        permission_data = {k: data[k] for k in ["enabled_repositories", "allowed_actions"] if k in data}
        if len(permission_data) > 0:
            status, body = await self.requester.async_request_raw(
                "PUT", f"/orgs/{org_id}/actions/permissions", json.dumps(permission_data)
            )

            if status != 204:
                raise RuntimeError(f"failed to update workflow settings for org '{org_id}'" f"\n{status}: {body}")

            print_debug(f"updated workflow settings for org '{org_id}'")

        if "selected_repository_ids" in data:
            await self._update_selected_repositories_for_workflow_settings(org_id, data["selected_repository_ids"])

        allowed_action_data = {
            k: data[k] for k in ["github_owned_allowed", "verified_allowed", "patterns_allowed"] if k in data
        }
        if len(allowed_action_data) > 0:
            await self._update_selected_actions_for_workflow_settings(org_id, allowed_action_data)

        default_permission_data = {
            k: data[k] for k in ["default_workflow_permissions", "can_approve_pull_request_reviews"] if k in data
        }
        if len(default_permission_data) > 0:
            await self._update_default_workflow_permissions(org_id, default_permission_data)

        print_debug(f"updated {len(data)} workflow setting(s)")

    async def _get_selected_repositories_for_workflow_settings(self, org_id: str) -> list[dict[str, Any]]:
        print_debug("retrieving selected repositories for org workflow settings")

        try:
            response = await self.requester.async_request_json(
                "GET", f"/orgs/{org_id}/actions/permissions/repositories"
            )
            return response["repositories"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}").with_traceback(tb)

    async def _update_selected_repositories_for_workflow_settings(
        self, org_id: str, selected_repository_ids: list[int]
    ) -> None:
        print_debug("updating selected repositories for org workflow settings")

        data = {"selected_repository_ids": selected_repository_ids}
        status, body = await self.requester.async_request_raw(
            "PUT", f"/orgs/{org_id}/actions/permissions/repositories", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(
                f"failed updating selected repositories for workflow settings of org '{org_id}'" f"\n{status}: {body}"
            )

        print_debug(f"updated selected repositories for workflow settings of org '{org_id}'")

    async def _get_selected_actions_for_workflow_settings(self, org_id: str) -> dict[str, Any]:
        print_debug(f"retrieving allowed actions for org '{org_id}'")

        try:
            return await self.requester.async_request_json(
                "GET", f"/orgs/{org_id}/actions/permissions/selected-actions"
            )
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving allowed actions for org '{org_id}':\n{ex}").with_traceback(tb)

    async def _update_selected_actions_for_workflow_settings(self, org_id: str, data: dict[str, Any]) -> None:
        print_debug(f"updating allowed actions for org '{org_id}'")

        status, body = await self.requester.async_request_raw(
            "PUT", f"/orgs/{org_id}/actions/permissions/selected-actions", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed updating allowed actions for org '{org_id}'" f"\n{status}: {body}")

        print_debug(f"updated allowed actions for org '{org_id}'")

    async def _get_default_workflow_permissions(self, org_id: str) -> dict[str, Any]:
        print_debug(f"retrieving default workflow permissions for org '{org_id}'")

        try:
            return await self.requester.async_request_json("GET", f"/orgs/{org_id}/actions/permissions/workflow")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving org default workflow permissions:\n{ex}").with_traceback(tb)

    async def _update_default_workflow_permissions(self, org_id: str, data: dict[str, Any]) -> None:
        print_debug(f"updating default workflow permissions for org '{org_id}'")

        status, body = await self.requester.async_request_raw(
            "PUT", f"/orgs/{org_id}/actions/permissions/workflow", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed updating default workflow permissions for org '{org_id}'" f"\n{status}: {body}")

        print_debug(f"updated default workflow permissions for org '{org_id}'")

    async def list_members(self, org_id: str, two_factor_disabled: bool) -> list[dict[str, Any]]:
        print_debug(f"retrieving list of organization members for org '{org_id}'")

        try:
            params = "?filter=2fa_disabled" if two_factor_disabled is True else ""
            return await self.requester.async_request_paged_json("GET", f"/orgs/{org_id}/members{params}")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving org default workflow permissions:\n{ex}").with_traceback(tb)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os.path
from abc import ABC, abstractmethod
from datetime import datetime
from functools import cached_property
from logging import Logger, getLogger
from typing import Generic, Optional, TypeVar, Union

from quart import current_app

from otterdog.config import OtterdogConfig
from otterdog.providers.github.auth import app_auth, token_auth
from otterdog.providers.github.rest import RestApi

_APP_REST_API: Optional[RestApi] = None
_INSTALLATION_REST_APIS: dict[str, tuple[RestApi, datetime]] = {}

_OTTERDOG_CONFIG: Optional[OtterdogConfig] = None

logger = getLogger(__name__)

T = TypeVar("T")


class Task(ABC, Generic[T]):
    @cached_property
    def logger(self) -> Logger:
        return getLogger(type(self).__name__)

    @staticmethod
    async def get_rest_api(installation_id: int) -> RestApi:
        return await get_rest_api_for_installation(installation_id)

    async def execute(self) -> Optional[T]:
        self.logger.debug(f"executing task '{self!r}'")

        await self._pre_execute()

        try:
            result = await self._execute()
            await self._post_execute(result)
            return result
        except RuntimeError as ex:
            self.logger.exception(f"failed to execute task '{self!r}'", exc_info=ex)
            await self._post_execute(ex)
            return None

    async def _pre_execute(self) -> None:
        pass

    async def _post_execute(self, result_or_exception: Union[T, Exception]) -> None:
        pass

    @abstractmethod
    async def _execute(self) -> T:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass


def _create_rest_api_for_app() -> RestApi:
    github_app_id = current_app.config["GITHUB_APP_ID"]
    github_app_private_key = current_app.config["GITHUB_APP_PRIVATE_KEY"]
    return RestApi(app_auth(github_app_id, github_app_private_key))


def get_rest_api_for_app() -> RestApi:
    global _APP_REST_API

    if _APP_REST_API is None:
        _APP_REST_API = _create_rest_api_for_app()

    return _APP_REST_API


async def get_rest_api_for_installation(installation_id: int) -> RestApi:
    global _INSTALLATION_REST_APIS
    installation = str(installation_id)

    current_api, expires_at = _INSTALLATION_REST_APIS.get(installation, (None, datetime.now()))
    if current_api is not None and expires_at is not None:
        if expires_at > datetime.now():
            return current_api

    token, expires_at = await get_rest_api_for_app().app.create_installation_access_token(installation)
    rest_api = RestApi(token_auth(token))
    _INSTALLATION_REST_APIS[installation] = (rest_api, expires_at)
    return rest_api


def get_otterdog_config() -> OtterdogConfig:
    global _OTTERDOG_CONFIG

    if _OTTERDOG_CONFIG is None:
        _OTTERDOG_CONFIG = _load_otterdog_config()

    return _OTTERDOG_CONFIG


def refresh_otterdog_config():
    global _OTTERDOG_CONFIG
    _OTTERDOG_CONFIG = _load_otterdog_config()


def _load_otterdog_config() -> OtterdogConfig:
    app_root = current_app.config["APP_ROOT"]
    config_file_url = current_app.config["OTTERDOG_CONFIG_URL"]

    logger.info(f"loading otterdog config from url '{config_file_url}'")

    import requests

    with requests.get(config_file_url) as response:
        config_file = os.path.join(app_root, "otterdog.json")
        with open(config_file, "w") as file:
            file.write(response.text)

    return OtterdogConfig(config_file, False, app_root)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
import re
from typing import Any, ClassVar, Optional, TypeVar, cast

from jsonbender import K, OptionalS, S, bend  # type: ignore

from otterdog.models import FailureType, ModelObject, ValidationContext
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    UNSET,
    associate_by_key,
    is_set_and_valid,
    is_unset,
    print_warn,
)

RS = TypeVar("RS", bound="Ruleset")


@dataclasses.dataclass
class Ruleset(ModelObject, abc.ABC):
    """
    Represents a Ruleset.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    name: str = dataclasses.field(metadata={"key": True})
    node_id: str = dataclasses.field(metadata={"external_only": True})
    enforcement: str

    bypass_actors: list[str]

    include_refs: list[str]
    exclude_refs: list[str]

    allows_creations: bool
    allows_deletions: bool
    allows_updates: bool
    allows_force_pushes: bool

    requires_commit_signatures: bool
    requires_linear_history: bool

    requires_pull_request: bool
    # the following settings are only taken into account
    # when requires_pull_request is True
    required_approving_review_count: Optional[int]
    dismisses_stale_reviews: bool
    requires_code_owner_review: bool
    requires_last_push_approval: bool
    requires_review_thread_resolution: bool

    requires_status_checks: bool
    requires_strict_status_checks: bool
    required_status_checks: list[str]

    requires_deployments: bool
    required_deployment_environments: list[str]

    _roles: ClassVar[dict[str, str]] = {"5": "RepositoryAdmin", "4": "Write", "2": "Maintain", "1": "OrganizationAdmin"}
    _inverted_roles: ClassVar[dict[str, str]] = dict((v, k) for k, v in _roles.items())

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        from .github_organization import GitHubOrganization

        org_settings = cast(GitHubOrganization, context.root_object).settings

        free_plan = org_settings.plan == "free"

        if is_set_and_valid(self.enforcement):
            if self.enforcement not in {"active", "disabled", "evaluate"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'enforcement' of value "
                    f"'{self.enforcement}', "
                    f"only values ('active' | 'disabled' | 'evaluate') are allowed.",
                )

            if free_plan is True and self.enforcement == "evaluate":
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'enforcement' of value "
                    f"'{self.enforcement}' which is not available for a 'free' plan.",
                )

        def valid_condition_pattern(pattern: str) -> bool:
            return pattern.startswith("refs/heads/") or ref == "~DEFAULT_BRANCH" or ref == "~ALL"

        if is_set_and_valid(self.include_refs):
            for ref in self.include_refs:
                if not valid_condition_pattern(ref):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has an invalid 'include_refs' pattern "
                        f"'{ref}', only values ('refs/heads/*', '~DEFAULT_BRANCH', '~ALL') are allowed",
                    )

        if is_set_and_valid(self.exclude_refs):
            for ref in self.exclude_refs:
                if not valid_condition_pattern(ref):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has an invalid 'exclude_refs' pattern "
                        f"'{ref}', only values ('refs/heads/*', '~DEFAULT_BRANCH', '~ALL') are allowed",
                    )

        if self.requires_pull_request is False:
            if is_set_and_valid(self.required_approving_review_count):
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has"
                    f" 'requires_pull_request' disabled but 'required_approving_review_count' "
                    f"is set to '{self.required_approving_review_count}', setting will be ignored.",
                )

            for key in {
                "dismisses_stale_reviews",
                "requires_code_owner_review",
                "requires_last_push_approval",
                "requires_review_thread_resolution",
            }:
                if self.__getattribute__(key) is True:
                    context.add_failure(
                        FailureType.INFO,
                        f"{self.get_model_header(parent_object)} has"
                        f" 'requires_pull_request' disabled but '{key}' "
                        f"is enabled, setting will be ignored.",
                    )

        # required_approving_review_count must be defined when requires_pull_request is enabled
        required_approving_review_count = self.required_approving_review_count
        if self.requires_pull_request is True and not is_unset(required_approving_review_count):
            if required_approving_review_count is None or required_approving_review_count < 0:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has"
                    f" 'requires_pull_request' enabled but 'required_approving_review_count' "
                    f"is not set (must be set to a non negative number).",
                )

        # if 'requires_status_checks' is disabled, issue a warning if required_status_checks is non-empty.
        required_status_checks = self.required_status_checks
        if (
            self.requires_status_checks is False
            and is_set_and_valid(required_status_checks)
            and len(required_status_checks) > 0
        ):
            context.add_failure(
                FailureType.INFO,
                f"{self.get_model_header(parent_object)} has"
                f" 'requires_status_checks' disabled but "
                f"'required_status_checks' is set to '{self.required_status_checks}', "
                f"setting will be ignored.",
            )

        # if 'requires_deployments' is disabled, issue a warning if required_deployment_environments is non-empty.
        if (
            self.requires_deployments is False
            and is_set_and_valid(self.required_deployment_environments)
            and len(self.required_deployment_environments) > 0
        ):
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header(parent_object)} has "
                f"'requires_deployments' disabled but "
                f"'required_deployment_environments' is set to "
                f"'{self.required_deployment_environments}', setting will be ignored.",
            )

        if self.requires_deployments is True and len(self.required_deployment_environments) > 0:
            from .repository import Repository

            environments = cast(Repository, parent_object).environments

            environments_by_name = associate_by_key(environments, lambda x: x.name)
            for env_name in self.required_deployment_environments:
                if env_name not in environments_by_name:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} requires deployment environment "
                        f"'{env_name}' which is not defined in the repository itself.",
                    )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        # disable diff computation for dependent fields of requires_pull_request,
        if self.requires_pull_request is False:
            if field.name in [
                "required_approving_review_count",
                "dismisses_stale_reviews",
                "requires_code_owner_review",
                "requires_last_push_approval",
                "requires_review_thread_resolution",
            ]:
                return False

        if self.requires_status_checks is False:
            if field.name in [
                "required_status_checks",
                "requires_strict_status_checks",
            ]:
                return False

        if self.requires_deployments is False:
            if field.name in ["required_deployment_environments"]:
                return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    @classmethod
    def from_model_data(cls, data: dict[str, Any]):
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]):
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping: dict[str, Any] = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}

        mapping.update(
            {
                "include_refs": OptionalS("conditions", "ref_name", "include", default=[]),
                "exclude_refs": OptionalS("conditions", "ref_name", "exclude", default=[]),
            }
        )

        rules = data.get("rules", [])

        def check_simple_rule(prop_key: str, rule_type: str, value_if_rule_is_present: bool) -> None:
            if any((_ := rule) for rule in rules if rule["type"] == rule_type):
                mapping[prop_key] = K(value_if_rule_is_present)
            else:
                mapping[prop_key] = K(not value_if_rule_is_present)

        check_simple_rule("allows_deletions", "deletion", False)
        check_simple_rule("allows_creations", "creation", False)
        check_simple_rule("allows_updates", "update", False)
        check_simple_rule("allows_force_pushes", "non_fast_forward", False)

        check_simple_rule("requires_commit_signatures", "required_signatures", True)
        check_simple_rule("requires_linear_history", "required_linear_history", True)

        # bypass actors
        if "bypass_actors" in data:
            bypass_actors = data["bypass_actors"]
            transformed_actors = []
            for actor in bypass_actors:
                actor_type = actor["actor_type"]

                if actor_type == "RepositoryRole" or actor_type == "OrganizationAdmin":
                    role = cls._roles.get(str(actor["actor_id"]), None)
                    if role is None:
                        print_warn(f"fail to map repository role '{actor['actor_id']}', skipping")
                        continue
                    transformed_actor = f"#{role}"
                elif actor_type == "Team":
                    transformed_actor = f"@{actor['team_slug']}"
                elif actor_type == "Integration":
                    transformed_actor = actor["app_slug"]
                else:
                    continue

                bypass_mode = actor.get("bypass_mode", "always")
                if bypass_mode != "always":
                    transformed_actor = f"{transformed_actor}:{bypass_mode}"

                transformed_actors.append(transformed_actor)

            mapping["bypass_actors"] = K(transformed_actors)

        else:
            mapping["bypass_actors"] = K([])

        # requires pull request
        if any((found := rule) for rule in rules if rule["type"] == "pull_request"):
            mapping["requires_pull_request"] = K(True)
            parameters = found.get("parameters", {})

            mapping["required_approving_review_count"] = K(parameters.get("required_approving_review_count", None))
            mapping["dismisses_stale_reviews"] = K(parameters.get("dismiss_stale_reviews_on_push", UNSET))
            mapping["requires_code_owner_review"] = K(parameters.get("require_code_owner_review", UNSET))
            mapping["requires_last_push_approval"] = K(parameters.get("require_last_push_approval", UNSET))
            mapping["requires_review_thread_resolution"] = K(parameters.get("required_review_thread_resolution", UNSET))
        else:
            mapping["requires_pull_request"] = K(False)
            mapping["required_approving_review_count"] = K(None)
            mapping["dismisses_stale_reviews"] = K(UNSET)
            mapping["requires_code_owner_review"] = K(UNSET)
            mapping["requires_last_push_approval"] = K(UNSET)
            mapping["requires_review_thread_resolution"] = K(UNSET)

        # required status checks
        if any((found := rule) for rule in rules if rule["type"] == "required_status_checks"):
            mapping["requires_status_checks"] = K(True)
            parameters = found.get("parameters", {})

            mapping["requires_strict_status_checks"] = K(parameters.get("strict_required_status_checks_policy", UNSET))

            raw_status_checks = parameters.get("required_status_checks", [])
            status_checks = []
            for status_check in raw_status_checks:
                if "app_slug" in status_check:
                    check = status_check["app_slug"] + ":" + status_check["context"]
                else:
                    check = status_check["context"]

                status_checks.append(check)

            mapping["required_status_checks"] = K(status_checks)
        else:
            mapping["requires_status_checks"] = K(False)
            mapping["requires_strict_status_checks"] = K(UNSET)
            mapping["required_status_checks"] = K([])

        # required deployments
        if any((found := rule) for rule in rules if rule["type"] == "required_deployments"):
            mapping["requires_deployments"] = K(True)
            deployment_environments = found.get("parameters", {}).get("required_deployment_environments", [])
            mapping["required_deployment_environments"] = K(deployment_environments)
        else:
            mapping["requires_deployments"] = K(False)
            mapping["required_deployment_environments"] = K([])

        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        def pop_mapping(keys: list[str]) -> None:
            for key_to_pop in keys:
                if key_to_pop in mapping:
                    mapping.pop(key_to_pop)

        # include / excludes
        ref_names = {}
        if "include_refs" in data:
            mapping.pop("include_refs")
            ref_names["include"] = K(data["include_refs"])

        if "exclude_refs" in data:
            mapping.pop("exclude_refs")
            ref_names["exclude"] = K(data["exclude_refs"])

        if len(ref_names) > 0:
            mapping["conditions"] = {"ref_name": ref_names}

        # bypass actors
        if "bypass_actors" in data:
            bypass_actors: list[str] = data["bypass_actors"]
            transformed_actors = []

            def extract_actor_and_bypass_mode(encoded_data: str) -> tuple[str, str]:
                if ":" in encoded_data:
                    a, m = re.split(":", encoded_data, 1)
                else:
                    a = encoded_data
                    m = "always"

                return a, m

            for actor in bypass_actors:
                if actor.startswith("#"):
                    role, bypass_mode = extract_actor_and_bypass_mode(actor[1:])
                    actor_type = "RepositoryRole"
                    actor_id = cls._inverted_roles[role]
                    if actor_id == "1":
                        actor_type = "OrganizationAdmin"
                elif actor.startswith("@"):
                    team, bypass_mode = extract_actor_and_bypass_mode(actor[1:])
                    actor_type = "Team"
                    actor_id = (await provider.rest_api.org.get_team_ids(team))[0]
                else:
                    app, bypass_mode = extract_actor_and_bypass_mode(actor)
                    actor_type = "Integration"
                    actor_id = (await provider.rest_api.app.get_app_ids(app))[0]

                transformed_actors.append(
                    {"actor_id": K(int(actor_id)), "actor_type": K(actor_type), "bypass_mode": K(bypass_mode)}
                )

            mapping["bypass_actors"] = transformed_actors

        # rules
        rules: list[Any] = []

        def add_simple_rule(prop_key: str, rule_type: str, inverted: bool) -> None:
            if prop_key in data:
                mapping.pop(prop_key)
                prop_value = data[prop_key]

                if (inverted and not prop_value) or (not inverted and prop_value):
                    rules.append({"type": K(rule_type)})

        add_simple_rule("allows_deletions", "deletion", True)
        add_simple_rule("allows_creations", "creation", True)
        add_simple_rule("allows_updates", "update", True)
        add_simple_rule("allows_force_pushes", "non_fast_forward", True)

        add_simple_rule("requires_commit_signatures", "required_signatures", False)
        add_simple_rule("requires_linear_history", "required_linear_history", False)

        def add_parameter(prop_key: str, param_key: str, params: dict[str, Any]):
            if prop_key in data:
                params[param_key] = S(prop_key)

        # requires pull request
        if "requires_pull_request" in data:
            value = data["requires_pull_request"]

            if value is True:
                rule = {"type": K("pull_request")}
                pull_parameters: dict[str, Any] = {}

                add_parameter("required_approving_review_count", "required_approving_review_count", pull_parameters)
                add_parameter("dismisses_stale_reviews", "dismiss_stale_reviews_on_push", pull_parameters)
                add_parameter("requires_code_owner_review", "require_code_owner_review", pull_parameters)
                add_parameter("requires_last_push_approval", "require_last_push_approval", pull_parameters)
                add_parameter("requires_review_thread_resolution", "required_review_thread_resolution", pull_parameters)

                rule["parameters"] = pull_parameters
                rules.append(rule)

        pop_mapping(
            [
                "requires_pull_request",
                "required_approving_review_count",
                "dismisses_stale_reviews",
                "requires_code_owner_review",
                "requires_last_push_approval",
                "requires_last_push_approval",
                "requires_review_thread_resolution",
            ]
        )

        # required status checks
        if "requires_status_checks" in data:
            value = data["requires_status_checks"]
            if value is True:
                rule = {"type": K("required_status_checks")}
                status_checks_parameters: dict[str, Any] = {}
                add_parameter(
                    "requires_strict_status_checks", "strict_required_status_checks_policy", status_checks_parameters
                )

                if "required_status_checks" in data:
                    required_status_checks = data["required_status_checks"]
                    app_slugs = set()

                    for check in required_status_checks:
                        if ":" in check:
                            app_slug, context = re.split(":", check, 1)

                            if app_slug != "any":
                                app_slugs.add(app_slug)

                    app_ids = await provider.get_app_ids(app_slugs)

                    transformed_checks = []
                    for check in required_status_checks:
                        if ":" in check:
                            app_slug, context = re.split(":", check, 1)

                            if app_slug == "any":
                                app_slug = None
                        else:
                            app_slug = None
                            context = check

                        if app_slug is None:
                            transformed_checks.append({"context": context})
                        else:
                            transformed_checks.append({"integration_id": app_ids[app_slug], "context": context})

                    status_checks_parameters["required_status_checks"] = K(transformed_checks)

                rule["parameters"] = status_checks_parameters
                rules.append(rule)

        pop_mapping(["requires_status_checks", "requires_strict_status_checks", "required_status_checks"])

        # # required deployments
        if "requires_deployments" in data:
            value = data["requires_deployments"]
            if value is True:
                rule = {"type": K("required_deployments")}
                deployment_parameters: dict[str, Any] = {}
                add_parameter(
                    "required_deployment_environments", "required_deployment_environments", deployment_parameters
                )
                rule["parameters"] = deployment_parameters
                rules.append(rule)

        pop_mapping(["requires_deployments", "required_deployment_environments"])

        if len(rules) > 0:
            mapping["rules"] = rules

        return mapping



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
import re
import subprocess
from typing import Any

from otterdog import utils
from otterdog.credentials import CredentialProvider, Credentials


class BitwardenVault(CredentialProvider):
    """
    A class to provide convenient access to a bitwarden vault.
    """

    def __init__(self, api_token_key: str):
        self._api_token_key = api_token_key

        utils.print_debug("unlocking bitwarden vault")
        self._status, output = subprocess.getstatusoutput("bw unlock --check")
        if self._status != 0:
            raise RuntimeError(f"could not access bitwarden vault:\n{output}")

        if not self.is_unlocked():
            raise RuntimeError("bitwarden vault is locked, run 'bw unlock' and follow instructions first")

    def is_unlocked(self) -> bool:
        return self._status == 0

    def get_credentials(self, org_name: str, data: dict[str, Any], only_token: bool = False) -> Credentials:
        assert self.is_unlocked()

        item_id = data.get("item_id")
        if item_id is None:
            raise RuntimeError("required key 'item_id' not found in authorization data")

        api_token_key = data.get("api_token_key")
        if api_token_key is None:
            api_token_key = self._api_token_key

        status, output = subprocess.getstatusoutput(f"bw get item {item_id}")
        if status != 0:
            raise RuntimeError(f"item with id '{item_id}' not found in your bitwarden vault: {output}")

        # load the item json string and access the field containing the GitHub token
        item = json.loads(output)

        token_field = next(filter(lambda k: k["name"] == api_token_key, item.get("fields", [])), None)
        if token_field is None:
            raise RuntimeError(f"field with key '{api_token_key}' not found in item with id '{item_id}'")

        github_token = token_field.get("value")
        if github_token is None:
            raise RuntimeError(f"field with key '{api_token_key}' is empty in item with id '{item_id}'")

        if only_token is False:
            username = item["login"]["username"]
            if username is None:
                raise RuntimeError(f"no username specified in item with id '{item_id}'")

            password = item["login"]["password"]
            if password is None:
                raise RuntimeError(f"no password specified in item with id '{item_id}'")

            totp_secret = item["login"]["totp"]
            if totp_secret is None:
                raise RuntimeError(f"totp is empty in item with id '{item_id}'")
        else:
            username = None
            password = None
            totp_secret = None

        return Credentials(username, password, totp_secret, github_token)

    def get_secret(self, data: str) -> str:
        try:
            item_id, secret_key = re.split("@", data)

            status, output = subprocess.getstatusoutput(f"bw get item {item_id}")
            if status != 0:
                raise RuntimeError(f"item with id '{item_id}' not found in your bitwarden vault: {output}")

            if status != 0:
                raise RuntimeError(f"item with id '{item_id}' not found in your bitwarden vault")

            # load the item json string and access the field containing the GitHub token
            item = json.loads(output)

            secret_field = next(filter(lambda k: k["name"] == secret_key, item.get("fields", [])), None)
            if secret_field is None:
                raise RuntimeError(f"field with key '{secret_key}' not found in item with id '{item_id}'")

            secret = secret_field.get("value")
            if secret is None:
                raise RuntimeError(f"field with key '{secret_key}' is empty in item with id '{item_id}'")

            return secret
        except ValueError:
            raise RuntimeError(f"failed to parse secret data '{data}'")

    def __repr__(self):
        return f"BitWardenVault(unlocked={self.is_unlocked()})"



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
import shutil

from otterdog.config import OrganizationConfig
from otterdog.models import PatchContext
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import get_approval, style

from . import Operation


class ImportOperation(Operation):
    """
    Imports the current live configuration for organizations.
    """

    def __init__(self, force_processing: bool, no_web_ui: bool):
        super().__init__()
        self._force_processing = force_processing
        self._no_web_ui = no_web_ui

    @property
    def force_processing(self) -> bool:
        return self._force_processing

    @property
    def no_web_ui(self) -> bool:
        return self._no_web_ui

    def pre_execute(self) -> None:
        self.printer.println("Importing resources:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")

        org_file_name = jsonnet_config.org_config_file

        if os.path.exists(org_file_name) and not self.force_processing:
            self.printer.println()
            self.printer.println(style("Definition already exists", bright=True) + f" at '{org_file_name}'.")
            self.printer.println("  Performing this action will overwrite its contents.")
            self.printer.println("  Do you want to continue? (Only 'yes' or 'y' will be accepted to approve)\n")

            self.printer.print(f"{style('Enter a value:', bright=True)} ")
            if not get_approval():
                self.printer.println("\nImport cancelled.")
                return 1

        if os.path.exists(org_file_name):
            sync_secrets_from_previous_config = True
            backup_file = f"{org_file_name}.bak"
            shutil.copy(org_file_name, backup_file)
            self.printer.println(f"\nExisting definition copied to '{style(backup_file, bright=True)}'.\n")
        else:
            sync_secrets_from_previous_config = False

        self.printer.level_up()

        try:
            try:
                credentials = self.config.get_credentials(org_config)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            if self.no_web_ui is True:
                self.printer.print_warn(
                    "the Web UI will not be queried as '--no-web-ui' has been specified, "
                    "the resulting config will be incomplete."
                )

            with GitHubProvider(credentials) as provider:
                organization = await GitHubOrganization.load_from_provider(
                    github_id, jsonnet_config, provider, self.no_web_ui, self.printer
                )

            # copy secrets from existing configuration if it is present.
            if sync_secrets_from_previous_config:
                self.printer.println("Copying secrets from previous configuration.")
                previous_organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
                organization.copy_secrets(previous_organization)

            context = PatchContext(github_id, organization.settings)
            output = organization.to_jsonnet(jsonnet_config, context)

            output_dir = jsonnet_config.org_dir
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)

            with open(org_file_name, "w") as file:
                file.write(output)

            self.printer.println(f"Organization definition written to '{org_file_name}'.")

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Any

from otterdog.providers.github.exception import GitHubException
from otterdog.utils import print_debug

from . import RestApi, RestClient


class PullRequestClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_pull_request(self, org_id: str, repo_name: str, pull_request_number: str) -> dict[str, Any]:
        print_debug(f"getting pull request with number '{pull_request_number}' from repo '{org_id}/{repo_name}'")

        try:
            return await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_request_number}"
            )
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving pull request:\n{ex}").with_traceback(tb)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import re
from typing import Any, Callable, ClassVar, Iterator, Optional, cast

from jsonbender import F, Forall, If, K, OptionalS, S, bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    LivePatchType,
    ModelObject,
    PatchContext,
    ValidationContext,
)
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    UNSET,
    Change,
    IndentingPrinter,
    is_set_and_present,
    is_set_and_valid,
    is_unset,
    write_patch_object_as_json,
)

from .branch_protection_rule import BranchProtectionRule
from .environment import Environment
from .organization_settings import OrganizationSettings
from .repo_ruleset import RepositoryRuleset
from .repo_secret import RepositorySecret
from .repo_variable import RepositoryVariable
from .repo_webhook import RepositoryWebhook
from .repo_workflow_settings import RepositoryWorkflowSettings


@dataclasses.dataclass
class Repository(ModelObject):
    """
    Represents a Repository of an Organization.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    node_id: str = dataclasses.field(metadata={"external_only": True})
    name: str = dataclasses.field(metadata={"key": True})
    description: Optional[str]
    homepage: Optional[str]
    private: bool
    has_discussions: bool
    has_issues: bool
    has_projects: bool
    has_wiki: bool
    is_template: bool
    template_repository: Optional[str] = dataclasses.field(metadata={"read_only": True})
    topics: list[str]
    default_branch: str
    allow_rebase_merge: bool
    allow_merge_commit: bool
    allow_squash_merge: bool
    allow_auto_merge: bool
    delete_branch_on_merge: bool
    allow_update_branch: bool
    squash_merge_commit_title: str
    squash_merge_commit_message: str
    merge_commit_title: str
    merge_commit_message: str
    archived: bool
    allow_forking: bool
    web_commit_signoff_required: bool
    secret_scanning: str
    secret_scanning_push_protection: str
    dependabot_alerts_enabled: bool
    dependabot_security_updates_enabled: bool

    gh_pages_build_type: str
    gh_pages_source_branch: Optional[str]
    gh_pages_source_path: Optional[str]

    forked_repository: Optional[str] = dataclasses.field(metadata={"model_only": True})
    fork_default_branch_only: bool = dataclasses.field(metadata={"model_only": True})

    workflows: RepositoryWorkflowSettings = dataclasses.field(metadata={"nested_model": True})

    # model only fields
    aliases: list[str] = dataclasses.field(metadata={"model_only": True}, default_factory=list)
    post_process_template_content: list[str] = dataclasses.field(metadata={"model_only": True}, default_factory=list)
    auto_init: bool = dataclasses.field(metadata={"model_only": True}, default=False)

    # nested model fields
    webhooks: list[RepositoryWebhook] = dataclasses.field(metadata={"nested_model": True}, default_factory=list)
    secrets: list[RepositorySecret] = dataclasses.field(metadata={"nested_model": True}, default_factory=list)
    variables: list[RepositoryVariable] = dataclasses.field(metadata={"nested_model": True}, default_factory=list)
    branch_protection_rules: list[BranchProtectionRule] = dataclasses.field(
        metadata={"nested_model": True}, default_factory=list
    )
    rulesets: list[RepositoryRuleset] = dataclasses.field(metadata={"nested_model": True}, default_factory=list)
    environments: list[Environment] = dataclasses.field(metadata={"nested_model": True}, default_factory=list)

    _security_properties: ClassVar[list[str]] = [
        "secret_scanning",
        "secret_scanning_push_protection",
        "dependabot_security_updates_enabled",
    ]

    _unavailable_fields_in_archived_repos: ClassVar[set[str]] = {
        "description",
        "homepage",
        "allow_auto_merge",
        "allow_merge_commit",
        "allow_rebase_merge",
        "allow_squash_merge",
        "allow_update_branch",
        "delete_branch_on_merge",
        "merge_commit_message",
        "merge_commit_title",
        "squash_merge_commit_message",
        "squash_merge_commit_title",
        "dependabot_alerts_enabled",
        "dependabot_security_updates_enabled",
        "secret_scanning",
        "secret_scanning_push_protection",
        "has_issues",
        "has_wiki",
        "has_projects",
    }

    _gh_pages_properties: ClassVar[list[str]] = [
        "gh_pages_source_branch",
        "gh_pages_source_path",
    ]

    @property
    def model_object_name(self) -> str:
        return "repository"

    def get_all_names(self) -> list[str]:
        return [self.name] + self.aliases

    def get_all_key_values(self) -> list[Any]:
        return self.get_all_names()

    def add_branch_protection_rule(self, rule: BranchProtectionRule) -> None:
        self.branch_protection_rules.append(rule)

    def set_branch_protection_rules(self, rules: list[BranchProtectionRule]) -> None:
        self.branch_protection_rules = rules

    def add_ruleset(self, rule: RepositoryRuleset) -> None:
        self.rulesets.append(rule)

    def set_rulesets(self, rules: list[RepositoryRuleset]) -> None:
        self.rulesets = rules

    def add_webhook(self, webhook: RepositoryWebhook) -> None:
        self.webhooks.append(webhook)

    def get_webhook(self, url: str) -> Optional[RepositoryWebhook]:
        return next(filter(lambda x: x.url == url, self.webhooks), None)  # type: ignore

    def set_webhooks(self, webhooks: list[RepositoryWebhook]) -> None:
        self.webhooks = webhooks

    def add_secret(self, secret: RepositorySecret) -> None:
        self.secrets.append(secret)

    def get_secret(self, name: str) -> Optional[RepositorySecret]:
        return next(filter(lambda x: x.name == name, self.secrets), None)  # type: ignore

    def set_secrets(self, secrets: list[RepositorySecret]) -> None:
        self.secrets = secrets

    def add_variable(self, variable: RepositoryVariable) -> None:
        self.variables.append(variable)

    def get_variable(self, name: str) -> Optional[RepositoryVariable]:
        return next(filter(lambda x: x.name == name, self.variables), None)  # type: ignore

    def set_variables(self, variables: list[RepositoryVariable]) -> None:
        self.variables = variables

    def add_environment(self, environment: Environment) -> None:
        self.environments.append(environment)

    def set_environments(self, environments: list[Environment]) -> None:
        self.environments = environments

    def coerce_from_org_settings(self, org_settings: OrganizationSettings) -> Repository:
        copy = dataclasses.replace(self)

        if org_settings.has_organization_projects is False:
            copy.has_projects = UNSET  # type: ignore

        if org_settings.web_commit_signoff_required is True:
            copy.web_commit_signoff_required = UNSET  # type: ignore

        return copy

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        from .github_organization import GitHubOrganization

        github_id = cast(GitHubOrganization, parent_object).github_id
        org_settings = cast(GitHubOrganization, parent_object).settings

        free_plan = org_settings.plan == "free"

        org_has_projects_disabled = org_settings.has_organization_projects is False
        org_web_commit_signoff_required = org_settings.web_commit_signoff_required is True
        org_members_cannot_fork_private_repositories = org_settings.members_can_fork_private_repositories is False

        org_has_discussions_enabled_with_repo_as_source = (
            org_settings.has_discussions is True
            and org_settings.discussion_source_repository == f"{github_id}/{self.name}"
        )

        is_private = self.private is True
        is_public = self.private is False

        allow_forking = self.allow_forking is True
        disallow_forking = self.allow_forking is False

        if is_set_and_present(self.description) and len(self.description) > 350:
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header()} has 'description' that exceeds the maximum allowed length of 350 chars.",
            )

        if is_set_and_valid(self.topics):
            if len(self.topics) > 20:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header()} has more than 20 'topics' defined.",
                )

            for topic in self.topics:
                if not self._valid_topic(topic):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header()} has defined an invalid topic '{topic}'. "
                        f"Only lower-case, numbers and '-' are allowed characters.",
                    )

        if is_public and disallow_forking:
            context.add_failure(
                FailureType.WARNING,
                f"public {self.get_model_header()} has 'allow_forking' disabled which is not permitted.",
            )

        has_wiki = self.has_wiki is True
        if is_private and has_wiki and free_plan:
            context.add_failure(
                FailureType.WARNING,
                f"private {self.get_model_header()} has 'has_wiki' enabled which"
                f"requires at least GitHub Team billing, "
                f"currently using '{org_settings.plan}' plan.",
            )

        has_discussions_disabled = self.has_discussions is False
        if has_discussions_disabled and org_has_discussions_enabled_with_repo_as_source:
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header()} has 'has_discussions' disabled "
                f"while the organization uses this repo as source repository for discussions.",
            )

        # it seems that 'has_repository_projects' is not really taken into account,
        # the setting 'has_organization_projects' is used to control whether repos can actually have projects.
        if self.has_projects is True and org_has_projects_disabled:
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header()} has 'has_projects' enabled "
                f"while the organization disables 'has_organization_projects', setting will be ignored.",
            )

        if is_private and org_members_cannot_fork_private_repositories and allow_forking:
            context.add_failure(
                FailureType.ERROR,
                f"private {self.get_model_header()} has 'allow_forking' enabled "
                f"while the organization disables 'members_can_fork_private_repositories'.",
            )

        if self.web_commit_signoff_required is False and org_web_commit_signoff_required:
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header()} has 'web_commit_signoff_required' disabled while "
                f"the organization requires it, setting will be ignored.",
            )

        secret_scanning_disabled = self.secret_scanning == "disabled"
        secret_scanning_push_protection_enabled = self.secret_scanning_push_protection == "enabled"
        if secret_scanning_disabled and secret_scanning_push_protection_enabled and self.archived is False:
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header()} has 'secret_scanning' disabled while "
                f"'secret_scanning_push_protection' is enabled.",
            )

        if is_set_and_valid(self.template_repository) and is_set_and_valid(self.forked_repository):
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header()} has 'template_repository' and 'forked_repository' set at the same time.",
            )

        for webhook in self.webhooks:
            webhook.validate(context, self)

        if self.archived is True:
            if len(self.branch_protection_rules) > 0:
                context.add_failure(
                    FailureType.INFO,
                    f"{self.get_model_header()} is archived but has branch_protection_rules, "
                    f"rules will be ignored.",
                )

        if self.name.lower() == f"{github_id}.github.io".lower():
            if self.gh_pages_build_type == "disabled":
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has"
                    f" 'gh_pages_build_type' disabled but the repo hosts the organization site and"
                    f" GitHub pages will be enabled automatically.\n"
                    f" Add the following snippet to your repo configuration:\n\n"
                    f'   gh_pages_build_type: "legacy",\n'
                    f'   gh_pages_source_branch: "main",\n'
                    f'   gh_pages_source_path: "/",',
                )

            if len(list(filter(lambda x: x.name == "github-pages", self.environments))) == 0:
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} hosts the organization site"
                    f" but no corresponding 'github-pages' environment, please add such an environment.\n"
                    f" Add the following snippet to the repo configuration:\n\n"
                    f"   environments: [\n"
                    f"     orgs.newEnvironment('github-pages') {{\n"
                    f"       branch_policies+: [\n"
                    f'         "main"\n'
                    f"       ],\n"
                    f'       deployment_branch_policy: "selected",\n'
                    f"     }},\n"
                    f"   ],",
                )
        elif is_set_and_valid(self.gh_pages_build_type):
            if self.gh_pages_build_type not in {"disabled", "legacy", "workflow"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"'gh_pages_build_type' has value '{self.gh_pages_build_type}', "
                    f"only values ('disabled' | 'legacy' | 'workflow') are allowed.",
                )

            if self.gh_pages_build_type == "disabled":
                for key in self._gh_pages_properties:
                    value = self.__getattribute__(key)
                    if value is not None:
                        context.add_failure(
                            FailureType.WARNING,
                            f"{self.get_model_header(parent_object)} has"
                            f" 'gh_pages_build_type' disabled but '{key}' "
                            f"is set to a value '{value}', setting will be ignored.",
                        )

            if self.gh_pages_build_type in {"legacy", "workflow"}:
                if len(list(filter(lambda x: x.name == "github-pages", self.environments))) == 0:
                    context.add_failure(
                        FailureType.WARNING,
                        f"{self.get_model_header(parent_object)} has"
                        f" 'gh_pages_build_type' with value '{self.gh_pages_build_type}' "
                        f"but no corresponding 'github-pages' environment, please add such an environment.",
                    )

        if is_set_and_valid(self.workflows):
            self.workflows.validate(context, self)

        for secret in self.secrets:
            secret.validate(context, self)

        for variable in self.variables:
            variable.validate(context, self)

        for bpr in self.branch_protection_rules:
            bpr.validate(context, self)

        for rule in self.rulesets:
            rule.validate(context, self)

        for env in self.environments:
            env.validate(context, self)

    @staticmethod
    def _valid_topic(topic, search=re.compile(r"[^a-z0-9\-]").search):
        return not bool(search(topic))

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        # private repos don't support security analysis.
        if self.private is True:
            if field.name in self._security_properties:
                return False

        if self.archived is True:
            if field.name in self._unavailable_fields_in_archived_repos:
                return False

        if self.gh_pages_build_type in ["disabled", "workflow"]:
            if field.name in self._gh_pages_properties:
                return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        # private repos don't support security analysis.
        if self.private is True:
            if field.name in self._security_properties:
                return False

        if self.gh_pages_build_type in ["disabled", "workflow"]:
            if field.name in self._gh_pages_properties:
                return False

        # when generating a patch, capture all the current configuration, even for
        # archived repos, the properties might be used when the repo gets unarchived.
        return True

    def get_model_objects(self) -> Iterator[tuple[ModelObject, ModelObject]]:
        for webhook in self.webhooks:
            yield webhook, self
            yield from webhook.get_model_objects()

        for secret in self.secrets:
            yield secret, self
            yield from secret.get_model_objects()

        for variable in self.variables:
            yield variable, self
            yield from variable.get_model_objects()

        for bpr in self.branch_protection_rules:
            yield bpr, self
            yield from bpr.get_model_objects()

        for ruleset in self.rulesets:
            yield ruleset, self
            yield from ruleset.get_model_objects()

        for env in self.environments:
            yield env, self
            yield from env.get_model_objects()

        if is_set_and_valid(self.workflows):
            yield self.workflows, self

    @classmethod
    def from_model_data(cls, data: dict[str, Any]) -> Repository:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}

        mapping.update(
            {
                "webhooks": OptionalS("webhooks", default=[]) >> Forall(lambda x: RepositoryWebhook.from_model_data(x)),
                "secrets": OptionalS("secrets", default=[]) >> Forall(lambda x: RepositorySecret.from_model_data(x)),
                "variables": OptionalS("variables", default=[])
                >> Forall(lambda x: RepositoryVariable.from_model_data(x)),
                "branch_protection_rules": OptionalS("branch_protection_rules", default=[])
                >> Forall(lambda x: BranchProtectionRule.from_model_data(x)),
                "rulesets": OptionalS("rulesets", default=[]) >> Forall(lambda x: RepositoryRuleset.from_model_data(x)),
                "environments": OptionalS("environments", default=[])
                >> Forall(lambda x: Environment.from_model_data(x)),
                "workflows": If(
                    OptionalS("workflows", default=None) == K(None),
                    K(UNSET),
                    S("workflows") >> F(lambda x: RepositoryWorkflowSettings.from_model_data(x)),
                ),
            }
        )

        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]) -> Repository:
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}

        def status_to_bool(status):
            if status == "enabled":
                return True
            elif status == "disabled":
                return False
            else:
                return UNSET

        # mapping for gh-pages config
        mapping.update(
            {
                "gh_pages_build_type": OptionalS("gh_pages", "build_type", default="disabled"),
                "gh_pages_source_branch": OptionalS("gh_pages", "source", "branch", default=None),
                "gh_pages_source_path": OptionalS("gh_pages", "source", "path", default=None),
            }
        )

        mapping.update(
            {
                "webhooks": K([]),
                "secrets": K([]),
                "variables": K([]),
                "branch_protection_rules": K([]),
                "rulesets": K([]),
                "environments": K([]),
                "secret_scanning": OptionalS("security_and_analysis", "secret_scanning", "status", default=UNSET),
                "secret_scanning_push_protection": OptionalS(
                    "security_and_analysis",
                    "secret_scanning_push_protection",
                    "status",
                    default=UNSET,
                ),
                "dependabot_security_updates_enabled": OptionalS(
                    "security_and_analysis", "dependabot_security_updates", "status", default=UNSET
                )
                >> F(status_to_bool),
                "template_repository": OptionalS("template_repository", "full_name", default=None),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        # add mapping for items that GitHub expects in a nested structure.

        # private repos do not support secret scanning settings, remove them.
        is_private = data.get("private", False)
        if is_private:
            for security_prop in cls._security_properties:
                if security_prop in mapping:
                    mapping.pop(security_prop)
        else:
            security_mapping = {}
            for security_prop in cls._security_properties:
                if security_prop in mapping:
                    mapping.pop(security_prop)
                if security_prop in data:
                    if security_prop.endswith("_enabled"):
                        github_security_prop = security_prop.removesuffix("_enabled")
                        security_mapping[github_security_prop] = {
                            "status": If(S(security_prop) == K(True), K("enabled"), K("disabled"))
                        }
                    else:
                        security_mapping[security_prop] = {"status": S(security_prop)}

            if len(security_mapping) > 0:
                mapping.update({"security_and_analysis": security_mapping})

        gh_pages_mapping = {}
        if "gh_pages_build_type" in data:
            mapping.pop("gh_pages_build_type")
            gh_pages_mapping["build_type"] = S("gh_pages_build_type")

        gh_pages_build_type = data.get("gh_pages_build_type", None)

        if gh_pages_build_type is None or gh_pages_build_type == "legacy":
            gh_pages_legacy_mapping = {}
            for source_prop in ["gh_pages_source_branch", "gh_pages_source_path"]:
                if source_prop in data:
                    mapping.pop(source_prop)
                    key = source_prop.rsplit("_")[-1]
                    gh_pages_legacy_mapping[key] = S(source_prop)

            if len(gh_pages_legacy_mapping) > 0:
                gh_pages_mapping["source"] = gh_pages_legacy_mapping

        if len(gh_pages_mapping) > 0:
            mapping["gh_pages"] = gh_pages_mapping

        return mapping

    def resolve_secrets(self, secret_resolver: Callable[[str], str]) -> None:
        for webhook in self.webhooks:
            webhook.resolve_secrets(secret_resolver)

        for secret in self.secrets:
            secret.resolve_secrets(secret_resolver)

    def copy_secrets(self, other_object: ModelObject) -> None:
        for webhook in self.webhooks:
            other_repo = cast(Repository, other_object)
            other_webhook = other_repo.get_webhook(webhook.url)
            if other_webhook is not None:
                webhook.copy_secrets(other_webhook)

        for secret in self.secrets:
            other_repo = cast(Repository, other_object)
            other_secret = other_repo.get_secret(secret.name)
            if other_secret is not None:
                secret.copy_secrets(other_secret)

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return f"orgs.{jsonnet_config.extend_repo}" if extend else f"orgs.{jsonnet_config.create_repo}"

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)

        has_webhooks = len(self.webhooks) > 0
        has_secrets = len(self.secrets) > 0
        has_variables = len(self.variables) > 0
        has_branch_protection_rules = len(self.branch_protection_rules) > 0
        has_rulesets = len(self.rulesets) > 0
        has_environments = len(self.environments) > 0

        # FIXME: take webhooks, branch protection rules and environments into account once
        #        it is supported for repos that get extended.
        has_changes = len(patch) > 0
        if extend and has_changes is False:
            return

        if "name" in patch:
            patch.pop("name")

        function = self.get_jsonnet_template_function(jsonnet_config, extend)
        printer.print(f"{function}('{self.name}')")

        write_patch_object_as_json(patch, printer, close_object=False)

        if is_set_and_present(self.workflows):
            default_workflow_settings = cast(Repository, default_object).workflows

            if is_set_and_valid(default_workflow_settings):
                coerced_settings = self.workflows.coerce_from_org_settings(
                    self, cast(OrganizationSettings, context.org_settings).workflows
                )
                patch = coerced_settings.get_patch_to(default_workflow_settings)
                if len(patch) > 0:
                    printer.print("workflows+:")
                    coerced_settings.to_jsonnet(printer, jsonnet_config, context, False, default_workflow_settings)

        # FIXME: support overriding webhooks for repos coming from the default configuration.
        if has_webhooks and not extend:
            default_repo_webhook = RepositoryWebhook.from_model_data(jsonnet_config.default_repo_webhook_config)

            printer.println("webhooks: [")
            printer.level_up()

            for webhook in self.webhooks:
                webhook.to_jsonnet(printer, jsonnet_config, context, False, default_repo_webhook)

            printer.level_down()
            printer.println("],")

        # FIXME: support overriding secrets for repos coming from the default configuration.
        if has_secrets and not extend:
            default_repo_secret = RepositorySecret.from_model_data(jsonnet_config.default_repo_secret_config)

            printer.println("secrets: [")
            printer.level_up()

            for secret in self.secrets:
                secret.to_jsonnet(printer, jsonnet_config, context, False, default_repo_secret)

            printer.level_down()
            printer.println("],")

        # FIXME: support overriding variables for repos coming from the default configuration.
        if has_variables and not extend:
            default_repo_variable = RepositoryVariable.from_model_data(jsonnet_config.default_repo_variable_config)

            printer.println("variables: [")
            printer.level_up()

            for variable in self.variables:
                variable.to_jsonnet(printer, jsonnet_config, context, False, default_repo_variable)

            printer.level_down()
            printer.println("],")

        # FIXME: support overriding branch protection rules for repos coming from
        #        the default configuration.
        if has_branch_protection_rules and not extend:
            default_org_rule = BranchProtectionRule.from_model_data(
                jsonnet_config.default_branch_protection_rule_config
            )

            printer.println("branch_protection_rules: [")
            printer.level_up()

            for rule in self.branch_protection_rules:
                rule.to_jsonnet(printer, jsonnet_config, context, False, default_org_rule)

            printer.level_down()
            printer.println("],")

        # FIXME: support overriding rulesets for repos coming from the default configuration.
        if has_rulesets and not extend:
            default_org_rule = RepositoryRuleset.from_model_data(jsonnet_config.default_repo_ruleset_config)

            printer.println("rulesets: [")
            printer.level_up()

            for ruleset in self.rulesets:
                ruleset.to_jsonnet(printer, jsonnet_config, context, False, default_org_rule)

            printer.level_down()
            printer.println("],")

        # FIXME: support overriding environments for repos coming from
        #        the default configuration.
        if has_environments and not extend:
            default_environment = Environment.from_model_data(jsonnet_config.default_environment_config)

            printer.println("environments: [")
            printer.level_up()

            for env in self.environments:
                env.to_jsonnet(printer, jsonnet_config, context, False, default_environment)

            printer.level_down()
            printer.println("],")

        # close the repo object
        printer.level_down()
        printer.println("},")

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: Optional[ModelObject],
        current_object: Optional[ModelObject],
        parent_object: Optional[ModelObject],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        if expected_object is None:
            assert isinstance(current_object, Repository)
            handler(LivePatch.of_deletion(current_object, parent_object, current_object.apply_live_patch))
            return

        assert isinstance(expected_object, Repository)

        expected_org_settings = cast(OrganizationSettings, context.expected_org_settings)
        coerced_object = expected_object.coerce_from_org_settings(expected_org_settings)

        if current_object is None:
            handler(LivePatch.of_addition(coerced_object, parent_object, coerced_object.apply_live_patch))
        else:
            assert isinstance(current_object, Repository)

            modified_repo: dict[str, Change[Any]] = coerced_object.get_difference_from(current_object)

            # FIXME: needed to add this hack to ensure that gh_pages_source_path is also present in
            #        the modified data as GitHub needs the path as well when the branch is changed.
            #        this needs to make clean to support making the diff operation generic as possible.
            if "gh_pages_source_branch" in modified_repo:
                gh_pages_source_path = cast(Repository, coerced_object).gh_pages_source_path
                modified_repo["gh_pages_source_path"] = Change(gh_pages_source_path, gh_pages_source_path)

            if len(modified_repo) > 0:
                handler(
                    LivePatch.of_changes(
                        coerced_object,
                        current_object,
                        modified_repo,
                        parent_object,
                        False,
                        coerced_object.apply_live_patch,
                    )
                )

        parent_repo = coerced_object if current_object is None else current_object

        if is_set_and_valid(coerced_object.workflows):
            RepositoryWorkflowSettings.generate_live_patch(
                coerced_object.workflows,
                current_object.workflows if current_object is not None else None,
                parent_repo,
                context,
                handler,
            )

        RepositoryWebhook.generate_live_patch_of_list(
            coerced_object.webhooks,
            current_object.webhooks if current_object is not None else [],
            parent_repo,
            context,
            handler,
        )

        RepositorySecret.generate_live_patch_of_list(
            coerced_object.secrets,
            current_object.secrets if current_object is not None else [],
            parent_repo,
            context,
            handler,
        )

        RepositoryVariable.generate_live_patch_of_list(
            coerced_object.variables,
            current_object.variables if current_object is not None else [],
            parent_repo,
            context,
            handler,
        )

        Environment.generate_live_patch_of_list(
            coerced_object.environments,
            current_object.environments if current_object is not None else [],
            parent_repo,
            context,
            handler,
        )

        # only take branch protection rules of non-archive projects into account
        if coerced_object.archived is False:
            BranchProtectionRule.generate_live_patch_of_list(
                coerced_object.branch_protection_rules,
                current_object.branch_protection_rules if current_object is not None else [],
                parent_repo,
                context,
                handler,
            )

            RepositoryRuleset.generate_live_patch_of_list(
                coerced_object.rulesets,
                current_object.rulesets if current_object is not None else [],
                parent_repo,
                context,
                handler,
            )

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        match patch.patch_type:
            case LivePatchType.ADD:
                assert isinstance(patch.expected_object, Repository)
                await provider.add_repo(
                    org_id,
                    await patch.expected_object.to_provider_data(org_id, provider),
                    patch.expected_object.template_repository,
                    patch.expected_object.post_process_template_content,
                    patch.expected_object.forked_repository,
                    patch.expected_object.fork_default_branch_only,
                    patch.expected_object.auto_init,
                )

            case LivePatchType.REMOVE:
                assert isinstance(patch.current_object, Repository)
                await provider.delete_repo(org_id, patch.current_object.name)

            case LivePatchType.CHANGE:
                assert patch.changes is not None
                assert isinstance(patch.expected_object, Repository)
                assert isinstance(patch.current_object, Repository)
                await provider.update_repo(
                    org_id,
                    patch.current_object.name,
                    await cls.changes_to_provider(org_id, patch.changes, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from otterdog.config import OrganizationConfig
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import associate_by_key, is_set_and_present, style

from . import Operation


class SyncTemplateOperation(Operation):
    """
    Syncs the contents of repositories created from a template repository.
    """

    def __init__(self, repo: str):
        super().__init__()
        self._repo = repo

    @property
    def repo(self) -> str:
        return self._repo

    def pre_execute(self) -> None:
        self.printer.println(f"Syncing organization repos '{self.repo}' from template master:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
        self.printer.level_up()

        try:
            org_file_name = jsonnet_config.org_config_file

            if not os.path.exists(org_file_name):
                self.printer.print_error(
                    f"configuration file '{org_file_name}' does not yet exist, run fetch-config or import first."
                )
                return 1

            try:
                organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
            except RuntimeError as ex:
                self.printer.print_error(f"failed to load configuration: {str(ex)}")
                return 1

            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                rest_api = provider.rest_api

                repositories_by_name = associate_by_key(organization.repositories, lambda r: r.name)
                repo = repositories_by_name.get(self.repo)
                if repo is not None and repo.archived is False:
                    if is_set_and_present(repo.template_repository):
                        self.printer.println(f'Syncing repository["{style(repo.name, bright=True)}"]')
                        updated_files = await rest_api.repo.sync_from_template_repository(
                            github_id,
                            repo.name,
                            repo.template_repository,
                            repo.post_process_template_content,
                        )

                        self.printer.level_up()
                        for file in updated_files:
                            self.printer.println(f"updated file '{file}'")
                        self.printer.level_down()

            return 0

        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Optional


class GitHubException(Exception):
    def __init__(self, url: Optional[str], status: int, data: str):
        self.__url = url
        self.__status = status
        self.__data = data

    @property
    def url(self) -> Optional[str]:
        return self.__url

    @property
    def status(self) -> int:
        return self.__status

    @property
    def data(self) -> str:
        return self.__data

    def __str__(self):
        return f"Exception while accessing {self.url}: (status={self.status}, body={self.data})"


class BadCredentialsException(GitHubException):
    def __str__(self):
        return f"Bad Credentials while accessing {self.url}: (body={self.data})"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
import textwrap
from io import StringIO

from otterdog.config import OrganizationConfig
from otterdog.models import ModelObject
from otterdog.models.github_organization import GitHubOrganization
from otterdog.models.repository import Repository
from otterdog.utils import IndentingPrinter, is_set_and_valid, style

from . import Operation


class ShowOperation(Operation):
    """
    Shows the local configuration of an organization.
    """

    def __init__(self, markdown: bool, output_dir: str):
        super().__init__()
        self._markdown = markdown
        self._output_dir = output_dir

    @property
    def markdown(self) -> bool:
        return self._markdown

    @property
    def output_dir(self) -> str:
        return self._output_dir

    def pre_execute(self) -> None:
        if not self.markdown:
            self.printer.println("Showing organization resources:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        if not self.markdown:
            self.printer.println(f"\nOrganization {style(org_config.name)}[id={github_id}]")
            self.printer.level_up()

        try:
            org_file_name = jsonnet_config.org_config_file

            if not os.path.exists(org_file_name):
                self.printer.print_error(
                    f"configuration file '{org_file_name}' does not yet exist, run 'fetch-config' or 'import first'"
                )
                return 1

            try:
                organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
            except RuntimeError as ex:
                self.printer.print_error(f"failed to load configuration: {str(ex)}")
                return 1

            if not self.markdown:
                self._print_classic(organization)
            else:
                self._print_markdown(organization)

            return 0

        finally:
            if not self.markdown:
                self.printer.level_down()

    def _print_classic(self, organization: GitHubOrganization) -> None:
        for model_object, parent_object in organization.get_model_objects():
            self.printer.println()
            model_header = model_object.get_model_header(parent_object)
            self.print_dict(model_object.to_model_dict(), model_header, "", "black")

    def _print_markdown(self, organization: GitHubOrganization) -> None:
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir, exist_ok=True)

        writer = StringIO()
        self.printer = IndentingPrinter(writer, spaces_per_level=4)

        self.printer.println(
            textwrap.dedent(
                """\
                ---
                hide:
                  - navigation
                ---
                """
            )
        )

        self.printer.println("# Current configuration")
        self.printer.println('=== "Organization Settings"')
        self.printer.level_up()
        self._print_model_object(organization.settings)
        self.printer.level_down()

        if is_set_and_valid(organization.settings.workflows):
            self.printer.println('=== "Workflow Settings"')
            self.printer.level_up()
            self._print_model_object(organization.settings.workflows)
            self.printer.level_down()

        self.printer.println('=== "Organization Webhooks"')
        self.printer.level_up()
        if len(organization.webhooks) > 0:
            self.printer.println("| URL | Uses SSL | Secret | Resolved Secret |")
            self.printer.println("| :-- | :------: | :----: | :-------------: |")

            for webhook in organization.webhooks:
                uses_ssl = ":white_check_mark:" if webhook.insecure_ssl == "0" else ":x:"
                has_secret = ":white_check_mark:" if is_set_and_valid(webhook.secret) else ":x:"
                resolved_secret = ":white_check_mark:" if not webhook.has_dummy_secret() else ":regional_indicator_x:"

                self.printer.println(f"| {webhook.url} | {uses_ssl} | {has_secret} | {resolved_secret} |")
        else:
            self.printer.println("No webhooks.")
        self.printer.level_down()

        self.printer.println('=== "Organization Secrets"')
        self.printer.level_up()
        if len(organization.secrets) > 0:
            self.printer.println("| Name | Resolved Secret |")
            self.printer.println("| :--- | :-------------: |")

            for secret in organization.secrets:
                resolved_secret = ":white_check_mark:" if not secret.has_dummy_secret() else ":regional_indicator_x:"

                self.printer.println(f"| {secret.name} | {resolved_secret} |")
        else:
            self.printer.println("No secrets.")
        self.printer.level_down()

        self.printer.println('=== "Organization Variables"')
        self.printer.level_up()
        if len(organization.variables) > 0:
            self.printer.println("| Name | Value |")
            self.printer.println("| :--- | :---- |")

            for variable in organization.variables:
                self.printer.println(f"| {variable.name} | {variable.value} |")
        else:
            self.printer.println("No variables.")
        self.printer.level_down()

        self.printer.println('=== "Repositories"')
        self.printer.level_up()

        self.printer.println("| Repository | Branch Protections | Secrets | Variables | Webhooks | Secret Scanning |")
        self.printer.println("| :--------- | :----------------: | :-----: | :-------: | :------: | :-------------: |")

        for repo in organization.repositories:
            has_branch_protections = (
                ":white_check_mark:" if len(repo.branch_protection_rules) > 0 or len(repo.rulesets) > 0 else ":x:"
            )
            has_secrets = ":white_check_mark:" if len(repo.secrets) > 0 else ":regional_indicator_x:"
            has_variables = ":white_check_mark:" if len(repo.variables) > 0 else ":regional_indicator_x:"
            has_webhooks = ":white_check_mark:" if len(repo.webhooks) > 0 else ":regional_indicator_x:"
            secret_scanning = ":white_check_mark:" if repo.secret_scanning == "enabled" else ":x:"

            label = ":material-archive:" if repo.archived else ""
            github_url = f"https://github.com/{organization.github_id}/{repo.name}"

            self.printer.println(
                f"| [{repo.name}](repo-{repo.name}.md) {label} "
                f"[:octicons-link-external-16:]({github_url}){{:target='_blank'}} | "
                f"{has_branch_protections} | {has_secrets} | {has_variables} | {has_webhooks} | "
                f"{secret_scanning} |"
            )

        self.printer.level_down()

        with open(os.path.join(self.output_dir, "configuration.md"), "w") as file:
            file.write(writer.getvalue())

        for repo in organization.repositories:
            self._print_repo_markdown(organization, repo)

    def _print_repo_markdown(self, organization: GitHubOrganization, repo: Repository) -> None:
        writer = StringIO()
        self.printer = IndentingPrinter(writer, spaces_per_level=4)

        self.printer.println(
            textwrap.dedent(
                """\
                ---
                hide:
                  - navigation
                ---
                """
            )
        )

        self.printer.println(f"# Repo [{repo.name}](https://github.com/{organization.github_id}/{repo.name})")
        self.printer.println('=== "Settings"')
        self.printer.level_up()
        self._print_model_object(repo)
        self.printer.level_down()

        if is_set_and_valid(repo.workflows):
            self.printer.println('=== "Workflows"')
            self.printer.level_up()
            self._print_model_object(repo.workflows)
            self.printer.level_down()

        self.printer.println('=== "Webhooks"')
        self.printer.level_up()

        if len(repo.webhooks) > 0:
            self.printer.println("| URL | Uses SSL | Secret | Resolved Secret |")
            self.printer.println("| :-- | :------: | :----: | :-------------: |")

            for webhook in repo.webhooks:
                uses_ssl = ":white_check_mark:" if webhook.insecure_ssl == "0" else ":x:"
                has_secret = ":white_check_mark:" if is_set_and_valid(webhook.secret) else ":x:"
                resolved_secret = ":white_check_mark:" if not webhook.has_dummy_secret() else ":regional_indicator_x:"

                self.printer.println(f"| {webhook.url} | {uses_ssl} | {has_secret} | {resolved_secret} |")
        else:
            self.printer.println("No webhooks.")
        self.printer.level_down()

        self.printer.println('=== "Secrets"')
        self.printer.level_up()

        if len(repo.secrets) > 0:
            self.printer.println("| Name | Resolved Secret |")
            self.printer.println("| :--- | :-------------: |")

            for secret in repo.secrets:
                resolved_secret = ":white_check_mark:" if not secret.has_dummy_secret() else ":regional_indicator_x:"

                self.printer.println(f"| {secret.name} | {resolved_secret} |")
        else:
            self.printer.println("No secrets.")
        self.printer.level_down()

        self.printer.println('=== "Variables"')
        self.printer.level_up()
        if len(repo.variables) > 0:
            self.printer.println("| Name | Value |")
            self.printer.println("| :--- | :---- |")

            for variable in repo.variables:
                self.printer.println(f"| {variable.name} | {variable.value} |")
        else:
            self.printer.println("No variables.")
        self.printer.level_down()

        self.printer.println('=== "Branch Protection Rules"')
        self.printer.level_up()
        if len(repo.branch_protection_rules) > 0:
            for bpr in repo.branch_protection_rules:
                self._print_model_object(bpr)
        else:
            self.printer.println("No branch protection rules.")
        self.printer.level_down()

        self.printer.println('=== "Rulesets"')
        self.printer.level_up()
        if len(repo.rulesets) > 0:
            for ruleset in repo.rulesets:
                self._print_model_object(ruleset)
        else:
            self.printer.println("No rulesets.")
        self.printer.level_down()

        with open(os.path.join(self.output_dir, f"repo-{repo.name}.md"), "w") as file:
            file.write(writer.getvalue())

    def _print_model_object(self, model_object: ModelObject, include_nested: bool = False):
        self.printer.println("``` jsonnet")
        self.print_dict(
            model_object.to_model_dict(include_nested_models=include_nested),
            model_object.model_object_name,
            "",
            "",
            ":",
            ",",
        )
        self.printer.println("```")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
import re
from typing import Any, Callable, Optional, TypeVar, cast

from jsonbender import OptionalS, S, bend  # type: ignore

from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    ModelObject,
    ValidationContext,
)
from otterdog.providers.github import GitHubProvider
from otterdog.utils import UNSET, Change, is_set_and_present, is_set_and_valid, is_unset

WT = TypeVar("WT", bound="Webhook")


@dataclasses.dataclass
class Webhook(ModelObject, abc.ABC):
    """
    Represents a Webhook.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    events: list[str]
    active: bool
    url: str = dataclasses.field(metadata={"key": True})
    content_type: str
    insecure_ssl: str
    secret: Optional[str]

    # model only fields
    aliases: list[str] = dataclasses.field(metadata={"model_only": True}, default_factory=list)

    def get_all_urls(self) -> list[str]:
        return [self.url] + self.aliases

    def get_all_key_values(self) -> list[Any]:
        return self.get_all_urls()

    def has_dummy_secret(self) -> bool:
        if is_set_and_valid(self.secret) and all(ch == "*" for ch in self.secret):  # type: ignore
            return True
        else:
            return False

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        return self.is_key_valid_for_diff_computation(field.name)

    def is_key_valid_for_diff_computation(self, key: str) -> bool:
        if key == "secret":
            return not self.has_dummy_secret()
        else:
            return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    def include_for_live_patch(self) -> bool:
        return not self.has_dummy_secret()

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if self.has_dummy_secret():
            context.add_failure(
                FailureType.INFO,
                f"{self.get_model_header(parent_object)} will be skipped during processing:\n"
                f"webhook has a secret set, but only a dummy secret '{self.secret}' is provided in "
                f"the configuration.",
            )
        elif is_set_and_present(self.secret) and ":" not in self.secret:
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header()} has a secret '{self.secret}' that does not use a credential provider.",
            )

        if is_set_and_valid(self.content_type):
            if self.content_type not in {"json", "form"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"'content_type' has value '{self.content_type}', " f"only values ('json' | 'form') are allowed.",
                )

        if is_set_and_valid(self.insecure_ssl):
            if self.insecure_ssl not in {"0", "1"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"'insecure_ssl' has value '{self.insecure_ssl}', " f"only values ('0' | '1') are allowed.",
                )

    @classmethod
    def from_model_data(cls, data: dict[str, Any]):
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]):
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        mapping.update(
            {
                "url": OptionalS("config", "url", default=UNSET),
                "content_type": OptionalS("config", "content_type", default=UNSET),
                "insecure_ssl": OptionalS("config", "insecure_ssl", default=UNSET),
                "secret": OptionalS("config", "secret", default=None),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        config_mapping = {}
        for config_prop in ["url", "content_type", "insecure_ssl", "secret"]:
            if config_prop in mapping:
                mapping.pop(config_prop)
                config_mapping[config_prop] = S(config_prop)

        if len(config_mapping) > 0:
            mapping["config"] = config_mapping

        return mapping

    def contains_secrets(self) -> bool:
        return self.secret is not None

    def resolve_secrets(self, secret_resolver: Callable[[str], str]) -> None:
        webhook_secret = self.secret
        if not is_unset(webhook_secret) and webhook_secret is not None:
            self.secret = secret_resolver(webhook_secret)

    def copy_secrets(self, other_object: ModelObject) -> None:
        if self.has_dummy_secret():
            self.secret = cast(Webhook, other_object).secret

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: Optional[ModelObject],
        current_object: Optional[ModelObject],
        parent_object: Optional[ModelObject],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        if current_object is None:
            assert isinstance(expected_object, Webhook)
            handler(LivePatch.of_addition(expected_object, parent_object, expected_object.apply_live_patch))
            return

        if expected_object is None:
            assert isinstance(current_object, Webhook)
            handler(LivePatch.of_deletion(current_object, parent_object, current_object.apply_live_patch))
            return

        assert isinstance(expected_object, Webhook)
        assert isinstance(current_object, Webhook)

        # if webhooks shall be updated and the webhook contains a valid secret perform a forced update.
        if (
            context.update_webhooks
            and is_set_and_valid(expected_object.secret)
            and re.match(context.update_filter, expected_object.url)
        ):
            model_dict = expected_object.to_model_dict()
            modified_webhook: dict[str, Change[Any]] = {k: Change(v, v) for k, v in model_dict.items()}

            handler(
                LivePatch.of_changes(
                    expected_object,
                    current_object,
                    modified_webhook,
                    parent_object,
                    True,
                    expected_object.apply_live_patch,
                )
            )
            return

        modified_webhook = expected_object.get_difference_from(current_object)

        if not is_unset(expected_object.secret):
            # special handling for secrets:
            #   if a secret was present by now its gone or vice-versa,
            #   include it in the diff view.
            expected_secret = expected_object.secret
            current_secret = current_object.secret

            def has_valid_secret(webhook: Webhook):
                return webhook.secret is not None and not webhook.has_dummy_secret()

            # if there are different unresolved secrets, display changes
            has_different_unresolved_secrets = (
                has_valid_secret(expected_object)
                and has_valid_secret(current_object)
                and expected_secret != current_secret
            )

            if (
                (expected_secret is not None and current_secret is None)
                or (expected_secret is None and current_secret is not None)
                or has_different_unresolved_secrets
            ):
                modified_webhook["secret"] = Change(current_secret, expected_secret)

        if len(modified_webhook) > 0:
            handler(
                LivePatch.of_changes(
                    expected_object,
                    current_object,
                    modified_webhook,
                    parent_object,
                    False,
                    expected_object.apply_live_patch,
                )
            )
